#!/usr/bin/env python2
# * **************************************************************** **
# File: parser.py
# Requires: Python 2.7+ (but not Python 3.0+)
# Note: For history, changes and dates for this file, consult git.
# Author: Brian Danilko, Likeable Software (brian@likeablesoftware.com)
# Copyright 2015-2017 Microbric Pty Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License (in the doc/licenses directory)
# for more details.
#
# * **************************************************************** */

""" Module to use python ast to parse an Ed.Py program, error check,
    and convert to an internal program representation. """

from __future__ import print_function
from __future__ import absolute_import

import ast
import re

# from . import util
from . import io
from . import program
from . import edpy_code

# ############ utility functions ########################################


def Name(node):
    className = str(node.__class__)
    match = re.match("\<class '_ast.([a-zA-Z]*)'\>.*", className)
    if match is None:
        io.Out.FatalRaw("ERROR - bad re")

    return match.group(1)


# def CheckTarget(node):
#     if ((Name(node) == "Name") or
#        ((Name(node) == "Attribute") and (Name(node.value) == "Name") and
#         (node.value.id == "self")) or CheckSlice(node)):
#         return
#     else:
#         io.Out.Error(io.TS.PARSE_SYNTAX_ERROR, "file:{0}:{1}: Syntax error",
#                      node.lineno, node.col_offset)
#         raise program.ParseError


def CheckCall(node):
    if (Name(node) != "Call"):
        raise RuntimeError("CheckCall() called with wrong argument")

    if ((len(node.keywords) == 0) and     # no keywords
        (getattr(node, "starargs", None) is None) and  # no starargs Py2
        (getattr(node, "kwargs", None) is None) and     # no kwargs Py2
        # TODO: check for Py3 starargs and kwargs
        GetVarName(node.func)):
        return GetVarName(node.func)
    else:
        io.Out.Error(io.TS.PARSE_TOO_COMPLEX,
                     "file:{0}:{1}: Syntax Error, {2} code too complex for Ed.Py",
                     node.lineno, node.col_offset, "CALL")
        raise program.ParseError


def GetVarName(node, noException=None):
    if (Name(node) == "Name"):
        return node.id
    elif ((Name(node) == "Attribute") and
          (Name(node.value) == "Name")):
        return node.value.id + "." + node.attr
    else:
        if noException is not None:
            return None
        else:
            io.Out.Error(io.TS.PARSE_TOO_COMPLEX,
                         "file:{0}:{1}: Syntax Error, {2} code too complex for Ed.Py",
                         node.lineno, node.col_offset, "CALL")
            raise program.ParseError


def CheckSlice(node):
    """Return True if a simple Subscript. If complex Subscript then throw."""
    nodeName = Name(node)
    if (nodeName == "Subscript"):
        # is it simple enough?
        if ((Name(node.value) != "Name") or     # only support a name for the slice variable
            (Name(node.slice) != "Index")):     # only support a simple index for the slice
                                                # though Index can be an expression

            io.Out.Error(io.TS.PARSE_TOO_COMPLEX,
                         "file:{0}:{1}: Syntax Error, {2} code too complex for Ed.Py",
                         node.lineno, node.col_offset, "ARRAY")
            raise program.ParseError
        return True
    return False


def CheckCompare(node):
    if (Name(node) != "Compare"):
        raise RuntimeError("CheckCompare() called with wrong argument")

    if ((len(node.ops) != 1) or     # no keywords
        (len(node.comparators) != 1)):

        io.Out.Error(io.TS.PARSE_TOO_COMPLEX,
                     "file:{0}:{1}: Syntax Error, {2} code too complex for Ed.Py",
                     node.lineno, node.col_offset, "COMPARE")
        raise program.ParseError

    if (Name(node.ops[0]) in ("In", "NotIn", "Is", "NotIs")):
        io.Out.Error(io.TS.PARSE_NOT_SUPPORTED,
                     "file:{0}:{1}: Syntax Error, {2}not supported in Ed.Py",
                     node.lineno, node.col_offset, "In/Is ")
        raise program.ParseError


def CheckWhile(node):
    # While(expr test, stmt* body, stmt* orelse)
    # Don't handle the orelse which is a 'else' clause
    if (Name(node) != "While"):
        raise RuntimeError("CheckWhile() called with wrong argument")

    if ((len(node.orelse) > 0)):             # no orelse
        io.Out.Error(io.TS.PARSE_TOO_COMPLEX,
                     "file:{0}:{1}: Syntax Error, {2} code too complex for Ed.Py",
                     node.lineno, node.col_offset, "WHILE")
        raise program.ParseError


def CheckFor(node):
    # For(expr target, expr iter, stmt* body, stmt* orelse)
    # target is the variable, iter is the array to be iterated over.
    # orelse is an 'else' clause which we don't handle
    if (Name(node) != "For"):
        raise RuntimeError("CheckFor() called with wrong argument")

    if ((len(node.orelse) > 0) or               # no orelse
        (Name(node.target) not in ("Name"))):  # want simple name

        io.Out.Error(io.TS.PARSE_TOO_COMPLEX,
                     "file:{0}:{1}: Syntax Error, {2} code too complex for Ed.Py",
                     node.lineno, node.col_offset, "FOR")
        raise program.ParseError

    if (Name(node.iter) == "Name"):
        return "Name"

    if ((Name(node.iter) != "Call") or (CheckCall(node.iter) != "range") or
        (len(node.iter.args) != 1) or
        ((Name(node.iter.args[0]) != "Num") and GetVarName(node.iter.args[0], False) is None)):
        io.Out.Error(io.TS.PARSE_TOO_COMPLEX,
                     "file:{0}:{1}: Syntax Error, {2} code too complex for Ed.Py",
                     node.lineno, node.col_offset, "FOR")
        raise program.ParseError

    return "Range"


def CheckNum(node):
    # verify that this is an integer
    if (Name(node) != "Num"):
        raise RuntimeError("CheckNum() called with wrong argument")

    # print("CheckNum with node:", node, "node.n:", node.n)

    if (isinstance(node.n, int) == False):
        io.Out.Error(io.TS.PARSE_CONST_NOT_INT,
                     "file:{0}:{1}: Syntax Error, constant {2} must be an integer value",
                     node.lineno, node.col_offset, node.n)
        raise program.ParseError

    # return them actual integer number
    return node.n


# ############ Class to convert from python ast to programIR  ################


class Converter(object):

    def __init__(self, programIR):
        self.program = programIR
        self.returnCode = 0
        self.ctlMarker = -1     # this value is pre-incremented before use
        self.loopStack = []     # used to associate break/continue with the enclosing while/for
        self.forIndex = 10000   # temp var used in for loops. Stay out of normal temp var way

    def WalkProgram(self, node):
        """Want to see a module node"""
        if (Name(node) != "Module"):
            io.Out.FatalRaw("Program needs to start with a Module node")

        # parse statements (and expressions for the doc strings)
        for s in node.body:
            name = Name(s)
            if (name == "FunctionDef"):
                # add a new function
                self.AddFunction(s)
            elif (name == "Import"):
                self.AddImport(s)
            elif (name == "ClassDef"):
                self.AddClass(s)
            else:
                self.AddFunctionStatement(self.program.Function["__main__"], s)

        return self.returnCode

    def WalkEdRoutines(self, node):
        """Walk the code for the Ed routines and add Ed. infront of the function names"""
        if (Name(node) != "Module"):
            io.Out.FatalRaw("Program needs to start with a Module node")

        # parse statements (and expressions for the doc strings)
        for s in node.body:
            name = Name(s)
            if (name == "FunctionDef"):
                # add a new function
                self.AddEdFunction(s)
            elif (name == "Import"):
                # ignore imports - so can have import Ed in code file
                pass
            elif (name == "ClassDef"):
                # ignore
                io.Out.FatalRaw("Ed internal functions must not have classes")
            else:
                # ignore anything not in a function
                pass

        return self.returnCode

    def AddEdFunction(self, node):
        """Add an Ed function to the program"""

        className = ""
        if (node.name.startswith("Ed_")):
            nodeName = "Ed." + node.name[3:]
        else:
            nodeName = node.name

        io.Out.DebugRaw("Adding Ed. function:", nodeName, node.__dict__)

        if (nodeName in self.program.Function):
            io.Out.Error(io.TS.PARSE_NAME_REUSED,
                         "file:{0}:{1}: Syntax Error, two {2} with the same name",
                         node.lineno, node.col_offset, "FUNCTIONS")
            raise program.ParseError

        if (len(node.decorator_list) > 0):
            io.Out.Error(io.TS.PARSE_NOT_SUPPORTED,
                         "file:{0}:{1}: Syntax Error, {2}not supported in Ed.Py",
                         node.lineno, node.col_offset, "DECORATORS ")
            raise program.ParseError

        try:
            paramNames = CheckGetParamNames(node)
        except:
            io.Out.Error(io.TS.PARSE_TOO_COMPLEX,
                         "file:{0}:{1}: Syntax Error, {2} code too complex for Ed.Py",
                         node.lineno, node.col_offset, "FUNCTION")
            raise program.ParseError
            
        
        # get all of the stuff in the function!
        newFunction = program.Function(nodeName, True)
        newFunction.args.extend(paramNames)

        # Enforce that all class methods first arg is always 'self'. Basically
        # good style and makes parsing and optimising easier.
        if (len(className) > 0):
            if ((len(newFunction.args) == 0) or (newFunction.args[0] != "self")):
                io.Out.Error(io.TS.PARSE_CLASS_ARG0_NOT_SELF,
                             "file:{0}:{1}: Syntax Error, first method arg must be 'self'",
                             node.lineno, node.col_offset)
                raise program.ParseError

        self.loopStack = []
        for s in node.body:
            self.AddFunctionStatement(newFunction, s)

        self.program.Function[nodeName] = newFunction

    def AddFunction(self, node, className=""):
        """Add a function to the program"""
        if (len(className) > 0):
            nodeName = className + "." + node.name
        else:
            nodeName = node.name

        io.Out.DebugRaw("Adding function:", nodeName, node.__dict__)

        if (nodeName in self.program.Function):
            io.Out.Error(io.TS.PARSE_NAME_REUSED,
                         "file:{0}:{1}: Syntax Error, two {2} with the same name",
                         node.lineno, node.col_offset, "FUNCTIONS")
            raise program.ParseError

        if (len(node.decorator_list) > 0):
            io.Out.Error(io.TS.PARSE_NOT_SUPPORTED,
                         "file:{0}:{1}: Syntax Error, {2}not supported in Ed.Py",
                         node.lineno, node.col_offset, "DECORATORS ")
            raise program.ParseError

        try:
            paramNames = CheckGetParamNames(node)
        except:
            io.Out.Error(io.TS.PARSE_TOO_COMPLEX,
                         "file:{0}:{1}: Syntax Error, {2} code too complex for Ed.Py",
                         node.lineno, node.col_offset, "FUNCTION")
            raise program.ParseError
        
        # get all of the stuff in the function!
        newFunction = program.Function(nodeName)
        newFunction.args.extend(paramNames)

        # Enforce that all class methods first arg is always 'self'. Basically
        # good style and makes parsing and optimising easier.
        if (len(className) > 0):
            if ((len(newFunction.args) == 0) or (newFunction.args[0] != "self")):
                io.Out.Error(io.TS.PARSE_CLASS_ARG0_NOT_SELF,
                             "file:{0}:{1}: Syntax Error, first method arg must be 'self'",
                             node.lineno, node.col_offset)
                raise program.ParseError

        self.loopStack = []
        for s in node.body:
            self.AddFunctionStatement(newFunction, s)

        self.program.Function[nodeName] = newFunction

    def AddFunctionStatement(self, function, node):
        name = Name(node)
        if (name in ("Assign", "AugAssign")):
            self.AddAssignStatement(function, node)
        elif (name == "Expr") and Name(node.value) == "Call":
            self.AddCallStatement(function, node.value)
        elif (name in ("While", "If")):
            # While(expr test, stmt* body, stmt* orelse)
            # If(expr test, stmt* body, stmt* orelse)

            if (name == "While"):
                CheckWhile(node)

            # line marker here to encompass the ControlMarker
            function.body.append(program.Marker(node.lineno, node.col_offset))

            # Create a control marker for this loop construct
            self.ctlMarker += 1
            markerNumber = self.ctlMarker

            # whileStack allows us to associate break/continue with the correct
            if (name == "While"):
                self.loopStack.append((name, markerNumber),)
                # print("LoopStack:", self.loopStack)

            function.body.append(program.ControlMarker(markerNumber, name, "start"))
            self.AddControlStatement(function, node, markerNumber)
            if (len(node.body) > 0):
                for l in node.body:
                    self.AddFunctionStatement(function, l)
            if (len(node.orelse) > 0):
                function.body.append(program.ControlMarker(markerNumber, name, "else"))
                for l in node.orelse:
                    self.AddFunctionStatement(function, l)
            function.body.append(program.ControlMarker(markerNumber, name, "end"))

            if (name == "While"):
                self.loopStack.pop()
                # print("LoopStack:", self.loopStack)

        elif (name in ("For")):
            # For(expr target, expr iter, stmt* body, stmt* orelse)
            # This is like a while loop but with assignment to an iterator
            # Only handle "for i in list" right now
            forType = CheckFor(node)

            # line marker here to encompass the ControlMarker
            function.body.append(program.Marker(node.lineno, node.col_offset))

            # Create a control marker for this loop construct
            self.ctlMarker += 1
            markerNumber = self.ctlMarker
            self.loopStack.append((name, markerNumber),)

            self.forIndex += 1
            forIndexTempNumber = self.forIndex
            forIndexValue = program.Value(name=forIndexTempNumber)

            # create a control value starting at -1 as we will preincrement

            function.body.append(program.UAssign(forIndexValue, "UAdd", program.Value(constant=-1)))

            function.body.append(program.ControlMarker(markerNumber, name, "start"))

            # increment forIndex
            function.body.append(program.BAssign(forIndexValue, forIndexValue, "Add",
                                                 program.Value(constant=1)))

            if (forType == "Range"):

                varName = GetVarName(node.iter.args[0], False)
                if (varName is not None):
                    limit = program.Value(name=varName)
                else:
                    limit = program.Value(constant=node.iter.args[0].n)

                # check that forIndexValue is in range of arrayName. If not then goto
                # the end control marker
                function.body.append(program.ForControl(markerNumber,
                                                        constantLimit=limit,
                                                        currentValue=forIndexValue))

                # set the value of the iterator
                function.body.append(program.UAssign(program.Value(name=node.target.id), "UAdd",
                                                     forIndexValue))

            else:
                arrayName = node.iter.id

                # check that forIndexValue is in range of arrayName. If not then goto
                # the end control marker
                function.body.append(program.ForControl(markerNumber,
                                                        arrayValue=program.Value(name=arrayName,
                                                                                 iVariable=forIndexTempNumber)))

                # set the value of the iterator
                function.body.append(program.UAssign(program.Value(name=node.target.id), "UAdd",
                                                     program.Value(name=arrayName,
                                                                   iVariable=forIndexTempNumber)))

            # the body of the for loop
            if (len(node.body) > 0):
                for l in node.body:
                    self.AddFunctionStatement(function, l)

            # This "for-end" will jump back to "for-start" with the same markerNumber
            function.body.append(program.ControlMarker(markerNumber, name, "end"))

            self.loopStack.pop()

        elif (name == "Global"):
            # Global has to be before any other executable statements in
            # the function (so easier for the students, and the compiler)!
            ok = (function.name != "__main__")
            if ok:
                for op in function.body:
                    if (op not in ("Marker", "Global")):
                        ok = False
                        break

            if not ok:
                io.Out.Error(io.TS.PARSE_GLOBAL_ORDER,
                             "file:{0}:{1}: Syntax error, globals must be first in functions",
                             node.lineno, node.col_offset)
                raise program.ParseError

            self.AddSimpleStatement(function, node)

        elif (name == "Return"):
            self.AddSimpleStatement(function, node)

        elif (name in ("Pass", "Break", "Continue")):
            if (name == "Pass"):
                pass   # drop "Pass" as it doesn't do anything
            else:
                # Simple program objects with the last while/for loop ctlMarker
                # to denote which control structure it modifies
                if (self.loopStack):
                    self.AddControlModifier(function, node, self.loopStack[-1][1])
                else:
                    io.Out.Error(io.TS.PARSE_NOT_IN_LOOP,
                                 "file:{0}:{1}: Syntax error, statement must be inside a loop",
                                 node.lineno, node.col_offset)
                    raise program.ParseError

        elif (name == "Expr") and Name(node.value) == "Str":
            function.docString = node.value.s
            io.Out.DebugRaw("Doc string:", node.value.s, function)
        else:
            io.Out.Error(io.TS.PARSE_INVALID_STATEMENT,
                         "file:{0}:{1}: Syntax error, statement not valid here",
                         node.lineno, node.col_offset)
            io.Out.DebugRaw("Invalid statement type:", name, node.__dict__)
            raise program.ParseError

    def AddAssignStatement(self, function, node):
        """Handle an assign or augmented assign, by reducing the ast
           embedded nodes to a list of assignments to temp variables."""

        line = node.lineno
        function.body.append(program.Marker(line, node.col_offset))
        name = Name(node)
        io.Out.DebugRaw("ASSIGN Statement:", name, node.__dict__)

        tempCount = 0
        statementList = []

        augOp = None
        if (name == "AugAssign"):
            augOp = Name(node.op)
            io.Out.DebugRaw("AugOp:", augOp)
            targetExpr = node.target
        elif (name == "Assign"):
            if (len(node.targets) != 1):
                io.Out.Error(io.TS.PARSE_SYNTAX_ERROR, "file:{0}:{1}: Syntax error",
                             node.lineno, node.offset)
                raise program.ParseError
            targetExpr = node.targets[0]

        # Evaluate the right hand side (the value)
        tempCount = self.HandleExpr(node.value, statementList, tempCount, line)
        # result of the right side of the = is now assigned to temp:0
        rhsStatements = len(statementList)

        # Evaluate the left hand side (the target) -- in case there is subscripting to be evaluated
        # The result of HandleExpr will be an UAssign() -- the operand of that will be the target
        tempCount = self.HandleExpr(targetExpr, statementList, tempCount, line)

        if ((len(statementList) == rhsStatements) or        # Make sure that evaluating target made
                                                            # a statement,
            (statementList[-1].kind != "UAssign") or        # and it is as expected.
            (statementList[-1].operation != "UAdd") or
            (statementList[-1].operand.IsConstant())):

            io.Out.Error(io.TS.PARSE_SYNTAX_ERROR, "file:{0}:{1}: Syntax error",
                         node.lineno, node.offset)
            raise program.ParseError

        target = statementList[-1].operand

        if (augOp is not None):
            # unrolling the AugAssign into a binary operation
            statementList[-1] = program.BAssign(target, target, augOp, program.Value(name=0))
        else:
            statementList[-1] = program.UAssign(target, "UAdd", program.Value(name=0))

        # Add the statementList to the function body
        for l in statementList:
            function.body.append(l)
            io.Out.DebugRaw("\t", l)

    def AddSimpleStatement(self, function, node):
        """Handle a return or a global"""

        line = node.lineno
        function.body.append(program.Marker(line, node.col_offset))
        name = Name(node)
        io.Out.DebugRaw("SIMPLE Statement:", name, node.__dict__)

        tempCount = 0
        statementList = []

        if (name == "Return"):
            if (node.value is None):
                if (function.returnsValue):
                    io.Out.Error(io.TS.PARSE_MIXED_RETURNS,
                                 "file:{0}:{1}: Syntax Error, all returns in a function must return a value or return nothing",
                         node.lineno, node.col_offset)
                    raise program.ParseError
                function.body.append(program.Return())
                function.returnsNone = True
            else:
                if (function.returnsNone):
                    io.Out.Error(io.TS.PARSE_MIXED_RETURNS,
                                 "file:{0}:{1}: Syntax Error, all returns in a function must return a value or return nothing",
                         node.lineno, node.col_offset)
                    raise program.ParseError
                returnTemp = tempCount
                self.HandleExpr(node.value, statementList, tempCount, line)

                for l in statementList:
                    function.body.append(l)
                    io.Out.DebugRaw("\t", l)

                function.body.append(program.Return(program.Value(name=returnTemp)))
                function.returnsValue = True
        else:
            # Add the global names directly into the function as it's before
            # any other statements in the function, order doesn't matter
            for n in node.names:
                if (n not in function.globalAccess):
                    function.globalAccess.append(n)

    def AddCallStatement(self, function, node):
        """Handle a 'call' by reducing the ast
           embedded nodes to a list of assignments to temp variables.
           Call is added here as it can be assigned to null if it is called
           outside a normal assign."""

        line = node.lineno
        function.body.append(program.Marker(line, node.col_offset))
        io.Out.DebugRaw("CALL Statement:", Name(node), node.__dict__)
        CheckCall(node)

        tempCount = 0
        statementList = []

        self.HandleExpr(node, statementList, tempCount, line)

        # the last line will be the call, assigned to variable:0. Remove the assignment

        if ((len(statementList) == 0) or                    # Make sure that evaluating target made
                                                            # a statement,
            (statementList[-1].kind != "Call")):            # and it is as expected.

            io.Out.Error(io.TS.PARSE_SYNTAX_ERROR, "file:{0}:{1}: Syntax error",
                         node.lineno, node.offset)
            raise program.ParseError

        statementList[-1].target = None

        for l in statementList:
            function.body.append(l)
            io.Out.DebugRaw("\t", l)

        # No need to add the final call as modified the last in the list

    def AddControlStatement(self, function, node, ctlMarker):
        """Handle an if or while statement - self.ctlMarker holds the  current loop"""

        line = node.lineno
        # line Marker was added by the caller

        name = Name(node)
        io.Out.DebugRaw("CONTROL Statement:", name, node.__dict__)

        tempCount = 0
        statementList = []

        # evaluate the test into a single variable
        tempCount = self.HandleExpr(node.test, statementList, tempCount, line)

        for l in statementList:
            function.body.append(l)
            io.Out.DebugRaw("\t", l)

        # add in the actual test code which uses the statements from the expression handler
        function.body.append(program.LoopControl(ctlMarker, name, program.Value(name=0)))

    def AddControlModifier(self, function, node, ctlMarker):
        """Handle a pass/break/continue statement - self.ctlMarker holds the  current loop"""

        line = node.lineno
        function.body.append(program.Marker(line, node.col_offset))
        name = Name(node)
        io.Out.DebugRaw("CONTROL Statement:", name, node.__dict__)

        function.body.append(program.LoopModifier(ctlMarker, name))

    def HandleExpr(self, node, statementList, tempCount, lineNo=None):
        """Convert expressions to assignments to tempCount in the statementList.
           Return the new tempCount"""

        # io.Out.DebugRaw("HandleExpr:{0}, list:{1}, tempCount:{2}".format(
        #     node, statementList, tempCount))

        nodeName = Name(node)
        target = program.Value(name=tempCount)  # Simple target to a temporary

        io.Out.DebugRaw("Value:", nodeName, node.__dict__)

        # Handle all terminals first

        if (nodeName == "Num"):
            CheckNum(node)
            # assign this to tempCount and return
            operand = program.Value(constant=node.n)
            statementList.append(program.UAssign(target, "UAdd", operand))
        elif (nodeName == "Name"):
            # assign this to tempCount and return
            operand = program.Value(name=node.id)
            statementList.append(program.UAssign(target, "UAdd", operand))
        elif (nodeName == "NameConstant"):
            # These nodes exist in Python 3
            operand = program.Value(name=str(node.value))
            statementList.append(program.UAssign(target, "UAdd", operand))
        elif (nodeName == "Attribute"):
            if (Name(node.value) == "Name"):
                operand = program.Value(name=node.value.id + "." + node.attr)
                statementList.append(program.UAssign(target, "UAdd", operand))
            else:
                io.Out.Error(io.TS.PARSE_NOT_SUPPORTED,
                             "file:{0}:{1}: Syntax Error, {2}not supported in Ed.Py",
                             lineNo, 0, "nested classes ")
                raise program.ParseError

        elif (nodeName == "Str"):
            # Only valid in Ed.TuneString() or ord() -- that error will be caught in the optimiser

            # if (len(node.s) != 1):
            #     io.Out.Error(io.TS.PARSE_NOT_SUPPORTED,
            #                  "file:{0}:{1}: Syntax Error, {2}not supported in Ed.Py",
            #                  lineNo, 0, "STRINGS ")
            #     raise program.ParseError

            # take the ord of the string, and assign this to tempCount and return
            operand = program.Value(strConst=node.s)
            statementList.append(program.UAssign(target, "UAdd", operand))

        elif (nodeName == "List"):
            # Only valid in Ed.List() -- that error will be caught in the optimiser

            # all of the elts must be just numbers! Should be able to handled vars, slices, functions,
            # etc. but too hard. Code can be written to initialise each value if the programmer wants.
            listInit = []
            for e in node.elts:
                if (Name(e) != "Num"):
                    io.Out.Error(io.TS.PARSE_TOO_COMPLEX,
                                 "file:{0}:{1}: Syntax Error, {2} code too complex for Ed.Py",
                                 node.lineno, node.col_offset, "LIST INIT")
                    raise program.ParseError

                CheckNum(e)
                listInit.append(e.n)

            operand = program.Value(listConst=listInit)
            statementList.append(program.UAssign(target, "UAdd", operand))

        elif (nodeName == "Subscript"):
            CheckSlice(node)
            # now know that the Name(node.value) == "Name" and Name(node.slice) == "Index"
            tempCount += 1
            operand = program.Value(name=node.value.id, iVariable=tempCount)
            # make another 1 or more assignment, return the next possible tempCount
            tempCount = self.HandleExpr(node.slice.value, statementList, tempCount, lineNo)
            statementList.append(program.UAssign(target, "UAdd", operand))

        elif (nodeName == "UnaryOp"):
            tempCount += 1
            operand = program.Value(name=tempCount)
            tempCount = self.HandleExpr(node.operand, statementList, tempCount, lineNo)
            statementList.append(program.UAssign(target, Name(node.op), operand))

        elif (nodeName == "BinOp"):
            op = Name(node.op)

            # do we support this operation?
            if (op in ("Pow")):
                io.Out.Error(io.TS.PARSE_NOT_SUPPORTED,
                             "file:{0}:{1}: Syntax Error, {2}not supported in Ed.Py",
                             lineNo, 0, "POWER ")
                raise program.ParseError

            # TODO: Evaluating left before right. Do we have to be more intelligent here?
            tempCount += 1
            left = program.Value(name=tempCount)
            tempCount = self.HandleExpr(node.left, statementList, tempCount, lineNo)
            tempCount += 1
            right = program.Value(name=tempCount)
            tempCount = self.HandleExpr(node.right, statementList, tempCount, lineNo)
            statementList.append(program.BAssign(target, left, op, right))

        elif (nodeName == "BoolOp"):
            # BoolOp(boolop op, expr* values)
            # BoolOp is strange as it can have a chain (more then 2) values,
            # including other BoolOps. So a or b or c and d has op OR, with values
            # (a, b, BoolOp(AND, (c, d)). Also it has short-circuit evaluation!

            # Algorithm: for values (a, b, c) and op OP:
            # 1. Bookmark this code with start and end markers with a nesting number
            # 2. Evaluate a
            # 3. Check if evaluation can be shortcircuited. If so set value and goto end
            # 4. Do 2 and 3 for all values
            # 5. If No shortcircuit, then set value to False for OR, True for AND, Done

            op = Name(node.op)
            self.ctlMarker += 1
            marker = self.ctlMarker

            resultTemp = tempCount

            # Mark start of BoolOp - name will be "And" or "Or"
            statementList.append(program.ControlMarker(marker, op, "start"))

            for v in node.values:
                tempCount += 1
                check = program.BoolCheck(marker, op, program.Value(name=tempCount),
                                          program.Value(name=resultTemp))
                tempCount = self.HandleExpr(v, statementList, tempCount, lineNo)
                statementList.append(check)

            if (op == "Or"):
                resultValue = program.Value(constant=0)
            else:
                resultValue = program.Value(constant=1)

            statementList.append(program.BoolCheck(marker, "Done", resultValue,
                                                   program.Value(name=resultTemp)))

            statementList.append(program.ControlMarker(marker, op, "end"))

        elif (nodeName == "Call"):
            # Call(expr func, expr* args, keyword* keywords, expr? starargs, expr? kwargs)
            # only func and args are valid
            funcName = CheckCall(node)
            # print (funcName)

            # Each arg can reuse temps -- in optimisation will check locality of
            # temps in a line and use the same temp if it doesn't overlap. But for
            # now, use different temps to keep the code simple
            args = []
            for a in node.args:
                tempCount += 1
                args.append(program.Value(name=tempCount))
                tempCount = self.HandleExpr(a, statementList, tempCount, lineNo)
            statementList.append(program.Call(target, funcName, args))

        elif (nodeName == "Compare"):
            # Compare(expr left, compop* ops, expr* comparators)
            CheckCompare(node)

            op = Name(node.ops[0])
            rhs = node.comparators[0]

            # TODO: Evaluating left before right. Do we have to be more intelligent here?
            tempCount += 1
            left = program.Value(name=tempCount)
            tempCount = self.HandleExpr(node.left, statementList, tempCount, lineNo)
            tempCount += 1
            right = program.Value(name=tempCount)
            tempCount = self.HandleExpr(rhs, statementList, tempCount, lineNo)
            statementList.append(program.BAssign(target, left, op, right))

        else:
            io.Out.Error(io.TS.PARSE_NOT_SUPPORTED,
                         "file:{0}:{1}: Syntax Error, {2}not supported in Ed.Py",
                         lineNo, 0, nodeName + " expr ")
            raise program.ParseError

        return tempCount

    def AddClass(self, node):
        """Add a function to the program"""
        io.Out.DebugRaw("Adding class:", node.name, node.__dict__)

        if (node.name in self.program.Class):
            io.Out.Error(io.TS.PARSE_NAME_REUSED,
                         "file:{0}:{1}: Syntax Error, two {2} with the same name",
                         node.lineno, node.col_offset, "CLASSES")

        if (len(node.decorator_list) > 0):
            io.Out.Error(io.TS.PARSE_NOT_SUPPORTED,
                         "file:{0}:{1}: Syntax Error, {2}not supported in Ed.Py",
                         node.lineno, node.col_offset, "DECORATORS ")
            raise program.ParseError

        if (len(node.bases) > 0):
            io.Out.Error(io.TS.PARSE_CLASS_NO_BASES_ALLOWED,
                         "file:{0}:{1}: Syntax Error, base classes are not allowed in Ed.Py",
                         node.lineno, node.col_offset)
            raise program.ParseError

        # Checks for using Ed as both a class and import is done in Opt
        newClass = program.Class(node.name)

        for e in node.body:
            name = Name(e)
            if (name == "FunctionDef"):
                # add a new function
                self.AddFunction(e, node.name)
                newClass.funcNames.append(e.name)
            elif (name == "Expr") and Name(e.value) == "Str":
                newClass.docString = e.value.s
                io.Out.DebugRaw("Doc string:", e.value.s, newClass)
            else:
                # only allow functions in class defintions
                io.Out.Error(io.TS.PARSE_CLASS_ALL_STATEMENTS_IN_FUNCTIONS,
                             "file:{0}:{1}: Syntax Error, in classes all statements must be in methods",
                             e.lineno, e.col_offset)
                raise program.ParseError

        self.program.Class[node.name] = newClass

    def AddImport(self, node):
        io.Out.DebugRaw("Adding import:", node.__dict__)
        if ((len(node.names) != 1) or (Name(node.names[0]) != "alias") or
            (node.names[0].asname is not None)):
            io.Out.Error(io.TS.PARSE_NOT_SUPPORTED,
                         "file:{0}:{1}: Syntax Error, {2}not supported in Ed.Py",
                         node.lineno, node.col_offset, "AS ")
            raise program.ParseError

        if ((len(self.program.Function) > 0 and "__main__" not in self.program.Function) or
            (len(self.program.Function) > 1) or (len(self.program.Class) > 0)):
            io.Out.Error(io.TS.PARSE_IMPORT_ORDER,
                         "file:{0}:{1}: Syntax error, imports must be before functions and classes",
                         node.lineno, node.col_offset)
            raise program.ParseError

        if (node.names[0].name != "Ed"):
            io.Out.Error(io.TS.PARSE_IMPORT_NOT_ED,
                         "file:{0}:{1}: Syntax error, only the Ed module can be imported",
                         node.lineno, node.col_offset)
            raise program.ParseError

        importName = node.names[0].name
        self.program.Import.append(importName)


def ConvertToIR(topNode, programIR, internalAst):
    c = Converter(programIR)

    try:
        rtc = c.WalkProgram(topNode)
        if (rtc == 0):
            # pass
            rtc = c.WalkEdRoutines(internalAst)

    except program.EdPyError:
        # Error was already raised
        rtc = 1
        if (io.Out.IsReRaiseSet()):
            raise

    except :
        rtc = 1
        io.Out.Error(io.TS.CMP_INTERNAL_ERROR,
                     "file::: Compiler internal error {0}", 701)
        if (io.Out.IsReRaiseSet()):
            raise

    return rtc


def Parse(filename, programIR):
    """Take a filename and a program IR and fill the program IR"""

    io.Out.Top(io.TS.PARSE_START, "Starting parse of file:{0}", filename)

    try:
        fh = open(filename, "rb")
        src = b"".join(fh.readlines())
        fh.close()
    except Exception:
        io.Out.Error(io.TS.FILE_OPEN_ERROR, "file:0: Could not access file {0}", filename)
        if (io.Out.IsReRaiseSet()):
            raise

        return 1

    return ParseString(src, filename, programIR)


def NormalPythonParse(programString, filename):
    """ Run python parser on the program to come up with an AST"""

    try:
        a = ast.parse(programString, filename)
    except SyntaxError as s:
        io.Out.Error(io.TS.PARSE_SYNTAX_ERROR, "file:{0}:{1}: Syntax error",
                     s.lineno, s.offset)
        if (io.Out.IsReRaiseSet()):
            raise
        return 2, None
    except TypeError:
        io.Out.Error(io.TS.BAD_INPUT_CHARS, "file:0: Illegal character in {0}", filename)
        if (io.Out.IsReRaiseSet()):
            raise
        return 2, None

    except Exception:
        io.Out.Error(io.TS.PARSE_ERROR, "file:0: There was an error parsing {0}", filename)
        if (io.Out.IsReRaiseSet()):
            raise
        return 2, None

    return 0, a


def ParseString(programString, filename, programIR):
    """Take programString (which is source lines ending in '\n'
       concatenated together) convert to the internal rep (IR)"""

    # First do the normal python parse of user's program
    error, ast = NormalPythonParse(programString, filename)
    if (error):
        return error

    # Now do the extra EdPy code which implements the Ed. functions
    src = edpy_code.CODE

    error, internalAst = NormalPythonParse(src, "INTERNAL_CODE")
    if (error):
        return error

    rtc = ConvertToIR(ast, programIR, internalAst)
    # io.Out.DebugRaw("Parse rtc:{}, ProgramIR:{}\n".format(rtc, programIR))

    if (io.Out.GetInfoDumpMask() & io.DUMP.PARSER):
        io.Out.DebugRaw("\nDump of internal representation after parsing (rtc:{0}):".format(rtc))
        programIR.Dump()
        io.Out.DebugRaw("\n")

    if (rtc != 0):
        io.Out.DebugRaw("WARNING - PARSER finished with an ERROR!!!\n")

    return rtc

def CheckGetParamNames(node):
    assert node.args.vararg == None
    assert node.args.kwarg == None
    assert len(node.args.defaults) == 0
    
    result = []
    for arg in node.args.args:
        if isinstance(arg, ast.Name): # Py2
            result.append(arg.id)
        elif isinstance(getattr(arg, "arg", None), str): # Py3
            result.append(arg.arg)
        else:
            raise Exception("Only simple parameters are supported")
    
    return result
    

# Only to be used as a module
if __name__ == '__main__':
    io.Out.FatalRaw("This file is a module and can not be run as a script!")



try:
    # Since Thonny 3.3
    from thonny.plugins.cpython.cpython_backend import get_backend
except ImportError:
    # Thonny 3.2
    from thonny.backend import get_vm as get_backend

import sys
import json
import os.path
import subprocess
import math
import time
import threading
import shutil

_completion_msg = "Done! Disconnect Edison and press its triangle button!"


def _compile_script(script_path):
    print("Compiling", script_path, "...")
    edpy_dir = os.path.join(os.path.dirname(__file__), "EdPy")
    edpy_script = os.path.join(edpy_dir, "EdPy.py")
    lang_file = os.path.join(edpy_dir, "en_lang.json")
    proc = subprocess.Popen(
        [sys.executable, edpy_script, lang_file, script_path],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    proc.wait()

    out = proc.stdout.read()
    err = proc.stderr.read()
    if err:
        print(err, file=sys.stderr)

    result = json.loads(out)

    if result["error"]:
        print()
        for msg in result["messages"]:
            print(_transform_error_msg(msg, script_path), file=sys.stderr)

        return None
    else:
        print("Compilation successful!")
        return result["wavFilename"]


def _transform_error_msg(msg, file_path):
    parts = list(map(str.strip, msg.split(":", maxsplit=4)))
    if (
        len(parts) == 5
        and parts[0] == "ERR"
        and parts[1] == "file"
        and parts[2].isnumeric()
    ):
        lineno = int(parts[2])
        if lineno > 0:
            return parts[-1] + '\n  File "{}", line {}'.format(file_path, lineno)
        else:
            return parts[-1]
    else:
        return msg


def program_edison(cmd):
    script_path = os.path.normpath(os.path.join(cmd["local_cwd"], cmd["args"][0]))
    wav_path = _compile_script(script_path)

    if wav_path:
        wav_length = _get_wav_duration(wav_path)
        print("Connect your Edison, turn volume to maximum and press the round button.")
        input("Press ENTER to start programming (takes %d seconds) " % wav_length)
        print("Programming ...")
        progress_thread = None
        completion_box = [None]
        tty_mode = cmd.get("tty_mode", False)
        if tty_mode:
            progress_thread = _show_progress_bar(
                math.ceil(wav_length), 50, completion_box
            )
        success = _play_wav(wav_path)
        completion_box[0] = success
        if progress_thread:
            progress_thread.join()

        if not tty_mode and success:
            print(_completion_msg)

        os.remove(wav_path)


def _show_progress_bar(seconds, length, completion_box):
    def work():
        print("Starting", end="")

        start_time = time.time()
        prev_completed_blocks = -1
        while True:
            if completion_box[0] is not None:
                break

            elapsed = time.time() - start_time
            progress = min(elapsed / seconds, 1)

            completed_blocks = int(progress * length)
            remaining_blocks = length - completed_blocks

            if completed_blocks != prev_completed_blocks:
                print(
                    "\r" + ("■" * completed_blocks) + ("□" * remaining_blocks), end=" "
                )
                prev_completed_blocks = completed_blocks

            time.sleep(0.1)

        if completion_box[0]:
            print("\r" + _completion_msg.ljust(length, " "))
        else:
            print("\r" + " ".ljust(length, " "))

    t = threading.Thread(target=work)
    t.start()
    return t


def _play_wav(path):
    if os.name == "nt":
        import winsound

        winsound.PlaySound(path, winsound.SND_FILENAME)
        return True
    elif sys.platform == "darwin":
        # http://stackoverflow.com/a/3498622/261181
        subprocess.check_call(["afplay", path])
        return True
    else:
        # https://www.techwalla.com/articles/how-to-play-a-wav-file-in-python
        if shutil.which("aplay"):
            subprocess.check_call(["aplay", "-q", path])
            return True
        else:
            print(
                "\nCommand 'aplay' not found. "
                + "Try installing 'alsa-utils' with your system package manager!",
                file=sys.stderr,
            )
            return False


def _get_wav_duration(path):
    # http://stackoverflow.com/questions/7833807/get-wav-file-length-or-duration
    import wave

    with wave.open(path, "r") as f:
        return f.getnframes() / f.getframerate()


def load_plugin():
    get_backend().add_command("ProgramEdison", program_edison)



#!/usr/bin/env python2
# * **************************************************************** **
# File: edpy_code.py
# Requires: Python 2.7+ (but not Python 3.0+)
# Note: For history, changes and dates for this file, consult git.
# Author: Brian Danilko, Likeable Software (brian@likeablesoftware.com)
# Copyright 2015-2017 Microbric Pty Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License (in the doc/licenses directory)
# for more details.
#
# * **************************************************************** */

"""Contains Ed.Py function python code. The compiler will compile
   this code as well to create a full program.
"""

# NOTE - to update comment out this line and the last one
# so that the code looks like normal python
CODE = """

# This is used to make the editor checking happy but is ignored
# by the EdPy parser!!
import Ed

# These routines will be accessed as if they were in the Ed module
# i.e. LeftLed() is accessed as Ed.LeftLed

# Some of the functions will be replaced in the compiler with direct
# code instead of calling the function -- basic inlining.


# Currently not used as the values are set in edpy_values
# def Ed_Init():
#     Ed.DistanceUnits = 0  # Ed.CM
#     Ed.Tempo = 1          # Ed.TEMPO_MEDIUM

# PYTHON (but not inline)

def abs(number):
    if (number < 0):
        return -number
    else:
        return number


# CONTROL

def Ed_LeftLed(value):
    value = value & 0x01
    Ed.WriteModuleRegister8Bit(Ed.MODULE_LEFT_LED, Ed.REG_LED_OUTPUT_8, value)


def Ed_RightLed(value):
    value = value & 0x01
    Ed.WriteModuleRegister8Bit(Ed.MODULE_RIGHT_LED, Ed.REG_LED_OUTPUT_8, value)


def Ed_ObstacleDetectionBeam(value):
    value = (value << 1) & 0x02    # bit 1 has the action!
    Ed.WriteModuleRegister8Bit(Ed.MODULE_IR_TX, Ed.REG_IRTX_ACTION_8, value)


def Ed_LineTrackerLed(value):
    value = value & 0x01
    Ed.WriteModuleRegister8Bit(Ed.MODULE_LINE_TRACKER, Ed.REG_LT_POWER_8, value)


def Ed_SendIRData(data):
    Ed.WriteModuleRegister8Bit(Ed.MODULE_IR_TX, Ed.REG_IRTX_ACTION_8, 0x00)
    Ed.WriteModuleRegister8Bit(Ed.MODULE_IR_TX, Ed.REG_IRTX_CHAR_8, data)
    Ed.WriteModuleRegister8Bit(Ed.MODULE_IR_TX, Ed.REG_IRTX_ACTION_8, 0x01)


def Ed_StartCountDown(time, units):
    # use the one-shot timer in the timer module, as it doesn't
    # stop execution
    units &= 0x01
    if (units == Ed.TIME_SECONDS):
        time *= 100   # convert from seconds to hundredths
    else:
        time /= 10    # convert from ms to hundredths
    Ed.WriteModuleRegister16Bit(Ed.MODULE_TIMERS, Ed.REG_TIMER_ONE_SHOT_16, time)
    Ed.WriteModuleRegister8Bit(Ed.MODULE_TIMERS, Ed.REG_TIMER_ACTION_8, 1)


def Ed_TimeWait(time, units):
    # use the pause timer in the timer module
    units &= 0x01
    if (units == Ed.TIME_SECONDS):
        time *= 100   # convert from seconds to hundredths
    else:
        time /= 10    # convert from ms to hundredths
    Ed.WriteModuleRegister16Bit(Ed.MODULE_TIMERS, Ed.REG_TIMER_PAUSE_16, time)
    Ed.WriteModuleRegister8Bit(Ed.MODULE_TIMERS, Ed.REG_TIMER_ACTION_8, 2)


def Ed_ResetDistance():
    # \QUESTION Gather that both motors reset distance?
    Ed.WriteModuleRegister16Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_DISTANCE_16, 0)
    Ed.WriteModuleRegister16Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_DISTANCE_16, 0)


def Ed_ReadObstacleDetection():
    mask = Ed.ReadModuleRegister8Bit(Ed.MODULE_IR_RX, Ed.REG_IRRX_STATUS_8)
    if (mask & Ed.OBSTACLE_DETECTED):
        if (mask & Ed.OBSTACLE_AHEAD):
            data = Ed.OBSTACLE_AHEAD
        else:
            data = mask & 0x38
        mask = mask & Ed.OBSTACLE_OTHER_MASK
        Ed.WriteModuleRegister8Bit(Ed.MODULE_IR_RX, Ed.REG_IRRX_STATUS_8, mask)
    else:
        data = 0

    return data


# MUSIC


def Ed_PlayBeep():
    Ed.WriteModuleRegister8Bit(Ed.MODULE_BEEPER, Ed.REG_BEEP_ACTION_8, 4)


def Ed_PlayMyBeep(freqCode):
    # doesn't use tempo - 50ms, so duration is 5
    Ed.WriteModuleRegister16Bit(Ed.MODULE_BEEPER, Ed.REG_BEEP_FREQ_16, freqCode)
    Ed.WriteModuleRegister16Bit(Ed.MODULE_BEEPER, Ed.REG_BEEP_DURATION_16, 5)
    Ed.WriteModuleRegister8Bit(Ed.MODULE_BEEPER, Ed.REG_BEEP_ACTION_8, 2)


def Ed_PlayTone(freqCode, durationMs):
    durationMs /= 10    # convert from ms to hundredths

    # doesn't use tempo
    Ed.WriteModuleRegister16Bit(Ed.MODULE_BEEPER, Ed.REG_BEEP_FREQ_16, freqCode)
    Ed.WriteModuleRegister16Bit(Ed.MODULE_BEEPER, Ed.REG_BEEP_DURATION_16, durationMs)
    Ed.WriteModuleRegister8Bit(Ed.MODULE_BEEPER, Ed.REG_BEEP_ACTION_8, 2)


def Ed_PlayTune(tuneString):
    addr = Ed.ObjectAddr(tuneString)

    # tempo has already been set
    Ed.WriteModuleRegister8Bit(Ed.MODULE_BEEPER, Ed.REG_BEEP_TUNE_STRING_8, addr)
    Ed.WriteModuleRegister8Bit(Ed.MODULE_BEEPER, Ed.REG_BEEP_ACTION_8, 8)


def Ed_ChangeTempo(newTempo):
    # set tempo first
    Ed.WriteModuleRegister16Bit(Ed.MODULE_BEEPER, Ed.REG_BEEP_TUNE_TEMPO_16, newTempo)

# MOVEMENT

# twoWheelArc = 226 # 36 * pi * 2
# oneWheelArc = 452 # 72 * pi * 2


def Ed_FinishDrive_SPACE(distance, left, right):
    while (distance > 0):
        distance = 0
        if (left != Ed.MOTOR_STOP_CODE):
            distance += Ed.ReadModuleRegister16Bit(Ed.MODULE_LEFT_MOTOR,
                                                   Ed.REG_MOTOR_DISTANCE_16)
        if (right != Ed.MOTOR_STOP_CODE):
            distance += Ed.ReadModuleRegister16Bit(Ed.MODULE_RIGHT_MOTOR,
                                                   Ed.REG_MOTOR_DISTANCE_16)


def Ed_FinishDrive_TIME(distance, left, right):
    if (distance > 0):
        # set up pause timer - when it completes the distance will be done
        distance /= 10    # convert from ms to hundredths
        Ed.WriteModuleRegister16Bit(Ed.MODULE_TIMERS, Ed.REG_TIMER_PAUSE_16, distance)
        Ed.WriteModuleRegister8Bit(Ed.MODULE_TIMERS, Ed.REG_TIMER_ACTION_8, 2)

        # turn off the motors
        if (left != Ed.MOTOR_STOP_CODE):
            Ed.WriteModuleRegister8Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_CONTROL_8,
                                       Ed.MOTOR_STOP_CODE)
        if (right != Ed.MOTOR_STOP_CODE):
            Ed.WriteModuleRegister8Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_CONTROL_8,
                                       Ed.MOTOR_STOP_CODE)


def Ed_Drive(direction, speed, distance):
    # placeholder
    pass


def Ed_Drive_CM(direction, speed, distance):
    if (direction < Ed.DIR_COMPLEX_START):
        Ed.DriveSimple_CM(direction, speed, distance, 1, 1)
    else:
        leftCtrl = Ed.MOTOR_STOP_CODE
        rightCtrl = Ed.MOTOR_STOP_CODE
        if (direction == Ed.FORWARD_RIGHT):
            leftCtrl = Ed.MOTOR_FOR_CODE
        elif (direction == Ed.BACKWARD_RIGHT):
            leftCtrl = Ed.MOTOR_BACK_CODE
        elif (direction == Ed.FORWARD_LEFT):
            rightCtrl = Ed.MOTOR_FOR_CODE
        elif (direction == Ed.BACKWARD_LEFT):
            rightCtrl = Ed.MOTOR_BACK_CODE
        elif (direction == Ed.SPIN_RIGHT):
            leftCtrl = Ed.MOTOR_FOR_CODE
            rightCtrl = Ed.MOTOR_BACK_CODE
        elif (direction == Ed.SPIN_LEFT):
            leftCtrl = Ed.MOTOR_BACK_CODE
            rightCtrl = Ed.MOTOR_FOR_CODE

        # distance is the degrees to rotate - zero degrees means unlimited
        if (distance != 0):

            # Let the robot turn at most 1 complete circle
            distance = distance % 360

            # As distance > 0, then it now being zero means that it was a multiple
            # of 360. In that case do one complete revolution.
            if (distance == 0):
                distance = 360

            # Distance in motor ticks is roughly equivalent to distance in degrees
            # just small correction for larger degrees
            if (distance > 300):
                distance += 2
            elif (distance > 100):
                distance += 1

            if (direction >= Ed.DIR_SPIN_START):
                distance /= 2

            if (distance == 0):
                distance = 1

            Ed.WriteModuleRegister16Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_DISTANCE_16, distance)
            Ed.WriteModuleRegister16Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_DISTANCE_16, distance)

            if (leftCtrl != Ed.MOTOR_STOP_CODE):
                leftCtrl |= Ed.MOTOR_DIST_CODE
            if (rightCtrl != Ed.MOTOR_STOP_CODE):
                rightCtrl |= Ed.MOTOR_DIST_CODE

        if (speed > Ed.SPEED_10):
            speed = Ed.SPEED_10

        Ed.WriteModuleRegister8Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_CONTROL_8, leftCtrl | speed)
        Ed.WriteModuleRegister8Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_CONTROL_8, rightCtrl | speed)

        Ed.FinishDrive_SPACE(distance, leftCtrl, rightCtrl)


def Ed_Drive_INCH(direction, speed, distance):
    if (direction < Ed.DIR_COMPLEX_START):
        Ed.DriveSimple_INCH(direction, speed, distance, 1, 1)
    else:
        leftCtrl = Ed.MOTOR_STOP_CODE
        rightCtrl = Ed.MOTOR_STOP_CODE
        if (direction == Ed.FORWARD_RIGHT):
            leftCtrl = Ed.MOTOR_FOR_CODE
        elif (direction == Ed.BACKWARD_RIGHT):
            leftCtrl = Ed.MOTOR_BACK_CODE
        elif (direction == Ed.FORWARD_LEFT):
            rightCtrl = Ed.MOTOR_FOR_CODE
        elif (direction == Ed.BACKWARD_LEFT):
            rightCtrl = Ed.MOTOR_BACK_CODE
        elif (direction == Ed.SPIN_RIGHT):
            leftCtrl = Ed.MOTOR_FOR_CODE
            rightCtrl = Ed.MOTOR_BACK_CODE
        elif (direction == Ed.SPIN_LEFT):
            leftCtrl = Ed.MOTOR_BACK_CODE
            rightCtrl = Ed.MOTOR_FOR_CODE

        # distance is the degrees to rotate - zero degrees means unlimited
        if (distance != 0):

            # Let the robot turn at most 1 complete circle
            distance = distance % 360

            # As distance > 0, then it now being zero means that it was a multiple
            # of 360. In that case do one complete revolution.
            if (distance == 0):
                distance = 360

            # Distance in motor ticks is roughly equivalent to distance in degrees
            # just small correction for larger degrees
            if (distance > 300):
                distance += 2
            elif (distance > 100):
                distance += 1

            if (direction >= Ed.DIR_SPIN_START):
                distance /= 2

            if (distance == 0):
                distance = 1

            Ed.WriteModuleRegister16Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_DISTANCE_16, distance)
            Ed.WriteModuleRegister16Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_DISTANCE_16, distance)

            if (leftCtrl != Ed.MOTOR_STOP_CODE):
                leftCtrl |= Ed.MOTOR_DIST_CODE
            if (rightCtrl != Ed.MOTOR_STOP_CODE):
                rightCtrl |= Ed.MOTOR_DIST_CODE

        if (speed > Ed.SPEED_10):
            speed = Ed.SPEED_10

        Ed.WriteModuleRegister8Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_CONTROL_8, leftCtrl | speed)
        Ed.WriteModuleRegister8Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_CONTROL_8, rightCtrl | speed)

        Ed.FinishDrive_SPACE(distance, leftCtrl, rightCtrl)


def Ed_Drive_TIME(direction, speed, distance):
    if (direction < Ed.DIR_COMPLEX_START):
        Ed.DriveSimple_TIME(direction, speed, distance, 1, 1)
    else:
        leftCtrl = Ed.MOTOR_STOP_CODE
        rightCtrl = Ed.MOTOR_STOP_CODE
        if (direction == Ed.FORWARD_RIGHT):
            leftCtrl = Ed.MOTOR_FOR_CODE
        elif (direction == Ed.BACKWARD_RIGHT):
            leftCtrl = Ed.MOTOR_BACK_CODE
        elif (direction == Ed.FORWARD_LEFT):
            rightCtrl = Ed.MOTOR_FOR_CODE
        elif (direction == Ed.BACKWARD_LEFT):
            rightCtrl = Ed.MOTOR_BACK_CODE
        elif (direction == Ed.SPIN_RIGHT):
            leftCtrl = Ed.MOTOR_FOR_CODE
            rightCtrl = Ed.MOTOR_BACK_CODE
        elif (direction == Ed.SPIN_LEFT):
            leftCtrl = Ed.MOTOR_BACK_CODE
            rightCtrl = Ed.MOTOR_FOR_CODE

        # for TIME, distance is the time in milliseconds

        if ((distance > 0) and (distance < 10)):
            # as distance is in milliseconds, but we convert to 1/100s of
            # of a second, if the distance is less then 10, we won't move at all
            return

        if (speed > Ed.SPEED_10):
            speed = Ed.SPEED_10

        Ed.WriteModuleRegister8Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_CONTROL_8, leftCtrl | speed)
        Ed.WriteModuleRegister8Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_CONTROL_8, rightCtrl | speed)

        Ed.FinishDrive_TIME(distance, leftCtrl, rightCtrl)


def Ed_DriveLeftMotor(direction, speed, distance):
    # placeholder
    pass


def Ed_DriveLeftMotor_CM(direction, speed, distance):
    if (direction <= Ed.BACKWARD):
        Ed.DriveSimple_CM(direction, speed, distance, 1, 0)


def Ed_DriveLeftMotor_INCH(direction, speed, distance):
    if (direction <= Ed.BACKWARD):
        Ed.DriveSimple_INCH(direction, speed, distance, 1, 0)


def Ed_DriveLeftMotor_TIME(direction, speed, distance):
    if (direction <= Ed.BACKWARD):
        Ed.DriveSimple_TIME(direction, speed, distance, 1, 0)


def Ed_DriveRightMotor(direction, speed, distance):
    # placeholder
    pass


def Ed_DriveRightMotor_CM(direction, speed, distance):
    if (direction <= Ed.BACKWARD):
        Ed.DriveSimple_CM(direction, speed, distance, 0, 1)


def Ed_DriveRightMotor_INCH(direction, speed, distance):
    if (direction <= Ed.BACKWARD):
        Ed.DriveSimple_INCH(direction, speed, distance, 0, 1)


def Ed_DriveRightMotor_TIME(direction, speed, distance):
    if (direction <= Ed.BACKWARD):
        Ed.DriveSimple_TIME(direction, speed, distance, 0, 1)


def Ed_DriveSimple_CM(direction, speed, distance, left, right):
    control = 0
    if (speed > Ed.SPEED_10):
        speed = Ed.SPEED_10

    if (direction == Ed.STOP):
        control = Ed.MOTOR_STOP_CODE
        distance = 0
    else:
        if (direction == Ed.FORWARD):
            control = Ed.MOTOR_FOR_CODE | speed
        else:
            control = Ed.MOTOR_BACK_CODE | speed

        if (distance > 0):
            distance *= 8
            # account for inertia by subtracting a fudge factor
            distance -= speed
            control = control | Ed.MOTOR_DIST_CODE

    if (left):
        Ed.WriteModuleRegister16Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_DISTANCE_16, distance)
        Ed.WriteModuleRegister8Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_CONTROL_8, control)

    if (right):
        Ed.WriteModuleRegister16Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_DISTANCE_16, distance)
        Ed.WriteModuleRegister8Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_CONTROL_8, control)

    Ed.FinishDrive_SPACE(distance, left, right)


def Ed_DriveSimple_INCH(direction, speed, distance, left, right):
    control = 0
    if (speed > Ed.SPEED_10):
        speed = Ed.SPEED_10

    if (direction == Ed.STOP):
        control = Ed.MOTOR_STOP_CODE
        distance = 0
    else:
        if (direction == Ed.FORWARD):
            control = Ed.MOTOR_FOR_CODE | speed
        else:
            control = Ed.MOTOR_BACK_CODE | speed

        if (distance > 0):
            # multiple by 20.3 in two steps
            distance *= 203
            distance /= 10
            # account for inertia by subtracting a fudge factor
            distance -= speed
            control = control | Ed.MOTOR_DIST_CODE

    if (left):
        Ed.WriteModuleRegister16Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_DISTANCE_16, distance)
        Ed.WriteModuleRegister8Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_CONTROL_8, control)

    if (right):
        Ed.WriteModuleRegister16Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_DISTANCE_16, distance)
        Ed.WriteModuleRegister8Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_CONTROL_8, control)

    Ed.FinishDrive_SPACE(distance, left, right)


def Ed_DriveSimple_TIME(direction, speed, distance, left, right):
    control = 0
    if (speed > Ed.SPEED_10):
        speed = Ed.SPEED_10

    if (direction == Ed.STOP):
        control = Ed.MOTOR_STOP_CODE
        distance = 0
    else:
        if (direction == Ed.FORWARD):
            control = Ed.MOTOR_FOR_CODE | speed
        else:
            control = Ed.MOTOR_BACK_CODE | speed

        if ((distance > 0) and (distance < 10)):
            # as distance is in milliseconds, but we convert to 1/100s of
            # of a second, if the distance is less then 10, we won't move at all
            return

    if (left):
        Ed.WriteModuleRegister8Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_CONTROL_8, control)

    if (right):
        Ed.WriteModuleRegister8Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_CONTROL_8, control)

    Ed.FinishDrive_TIME(distance, left, right)


def Ed_SetDistance(which, distance):
    # placeholder
    pass


def Ed_SetDistance_CM(which, distance):
    if (distance > 0):
        distance *= 8

        if ((which & 0x01) == Ed.MOTOR_LEFT):
            Ed.WriteModuleRegister16Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_DISTANCE_16, distance)
            # Turn on the 5th bit, which turns on distance checking
            Ed.SetModuleRegisterBit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_CONTROL_8, 5)
        else:
            Ed.WriteModuleRegister16Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_DISTANCE_16, distance)
            # Turn on the 5th bit, which turns on distance checking
            Ed.SetModuleRegisterBit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_CONTROL_8, 5)


def Ed_SetDistance_INCH(which, distance):
    if (distance > 0):
        # multiple by 20.3 in two steps
        distance *= 203
        distance /= 10

        if ((which & 0x01) == Ed.MOTOR_LEFT):
            Ed.WriteModuleRegister16Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_DISTANCE_16, distance)
            # Turn on the 5th bit, which turns on distance checking
            Ed.SetModuleRegisterBit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_CONTROL_8, 5)
        else:
            Ed.WriteModuleRegister16Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_DISTANCE_16, distance)
            # Turn on the 5th bit, which turns on distance checking
            Ed.SetModuleRegisterBit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_CONTROL_8, 5)


# READ DATA


def Ed_ReadKeypad():
    # TODO: Do we have to clear this register??
    button = Ed.ReadModuleRegister8Bit(Ed.MODULE_DEVICES, Ed.REG_DEV_BUTTON_8)
    Ed.WriteModuleRegister8Bit(Ed.MODULE_DEVICES, Ed.REG_DEV_BUTTON_8, 0)
    return button & Ed.KEYPAD_MASK   # Assuming a bitmask


def Ed_ReadRandom():
    Ed.WriteModuleRegister8Bit(Ed.MODULE_DEVICES, Ed.REG_DEV_ACTION_8, 0x10)
    return Ed.ReadModuleRegister8Bit(Ed.MODULE_DEVICES, Ed.REG_DEV_RANDOM_8)


def Ed_ReadClapSensor():
    data = (Ed.ReadModuleRegister8Bit(Ed.MODULE_BEEPER, Ed.REG_BEEP_STATUS_8) & Ed.CLAP_MASK)
    if data:
        Ed.ClearModuleRegisterBit(Ed.MODULE_BEEPER, Ed.REG_BEEP_STATUS_8, Ed.CLAP_DETECTED_BIT)
    return data


def Ed_ReadLineState():
    return (Ed.ReadModuleRegister8Bit(Ed.MODULE_LINE_TRACKER, Ed.REG_LT_STATUS_8) & Ed.LINE_MASK)


def Ed_ReadLineChange():
    change = Ed.ReadModuleRegister8Bit(Ed.MODULE_LINE_TRACKER, Ed.REG_LT_STATUS_8) & Ed.LINE_CHANGE_MASK
    if (change):
        Ed.ClearModuleRegisterBit(Ed.MODULE_LINE_TRACKER, Ed.REG_LT_STATUS_8, Ed.LINE_CHANGE_BIT)
        return 1
    else:
        return 0


def Ed_ReadRemote():
    if ((Ed.ReadModuleRegister8Bit(Ed.MODULE_IR_RX, Ed.REG_IRRX_STATUS_8) & 0x02) == 0):
        return Ed.REMOTE_CODE_NONE

    data = Ed.ReadModuleRegister8Bit(Ed.MODULE_IR_RX, Ed.REG_IRRX_MATCH_INDEX_8)
    Ed.ClearModuleRegisterBit(Ed.MODULE_IR_RX, Ed.REG_IRRX_STATUS_8, 1)
    return data


def Ed_ReadIRData():
    data = Ed.ReadModuleRegister8Bit(Ed.MODULE_IR_RX, Ed.REG_IRRX_RCV_CHAR_8)
    Ed.WriteModuleRegister8Bit(Ed.MODULE_IR_RX, Ed.REG_IRRX_RCV_CHAR_8, 0)
    Ed.ClearModuleRegisterBit(Ed.MODULE_IR_RX, Ed.REG_IRRX_STATUS_8, 0)
    return data


def Ed_ReadLeftLightLevel():
    return Ed.ReadModuleRegister16Bit(Ed.MODULE_LEFT_LED, Ed.REG_LED_LEVEL_16)


def Ed_ReadRightLightLevel():
    return Ed.ReadModuleRegister16Bit(Ed.MODULE_RIGHT_LED, Ed.REG_LED_LEVEL_16)


def Ed_ReadLineTracker():
    return Ed.ReadModuleRegister16Bit(Ed.MODULE_LINE_TRACKER, Ed.REG_LT_LEVEL_16)


def Ed_ReadCountDown(units):
    time = Ed.ReadModuleRegister16Bit(Ed.MODULE_TIMERS, Ed.REG_TIMER_ONE_SHOT_16)
    if ((units & 0x01) == Ed.TIME_SECONDS):
        time /= 100   # convert from hundredths to seconds
    else:
        time *= 10    # convert from hundredths to ms
    return time


def Ed_ReadMusicEnd():
    # end if tune done, tone done
    result = 0
    status = Ed.ReadModuleRegister8Bit(Ed.MODULE_BEEPER, Ed.REG_BEEP_STATUS_8)
    # don't clear clap detected status or tune error
    if (status & 0x01):
        Ed.ClearModuleRegisterBit(Ed.MODULE_BEEPER, Ed.REG_BEEP_STATUS_8, 0)
        result = 1
    if (status & 0x02):
        Ed.ClearModuleRegisterBit(Ed.MODULE_BEEPER, Ed.REG_BEEP_STATUS_8, 1)
        result = 1
    return result

def Ed_ReadTuneError():
    # Don't clear anything, just return 1 if error bit set
    return ((Ed.ReadModuleRegister8Bit(Ed.MODULE_BEEPER, Ed.REG_BEEP_STATUS_8) & 8) != 0)


def Ed_ReadDriveLoad():
    # OR the values from both wheels
    value = (Ed.ReadModuleRegister8Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_STATUS_8) &
             Ed.DRIVE_STRAINED)
    value |= (Ed.ReadModuleRegister8Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_STATUS_8) &
              Ed.DRIVE_STRAINED)
    return value


def Ed_ReadDistance(which):
    # placeholder
    pass


def Ed_ReadDistance_CM(which):
    if ((which & 0x01) == Ed.MOTOR_LEFT):
        which = Ed.ReadModuleRegister16Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_DISTANCE_16)
    else:
        which = Ed.ReadModuleRegister16Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_DISTANCE_16)

    # Convert to the correct units - CM
    which /= 8

    return which


def Ed_ReadDistance_INCH(which):
    if ((which & 0x01) == Ed.MOTOR_LEFT):
        which = Ed.ReadModuleRegister16Bit(Ed.MODULE_LEFT_MOTOR, Ed.REG_MOTOR_DISTANCE_16)
    else:
        which = Ed.ReadModuleRegister16Bit(Ed.MODULE_RIGHT_MOTOR, Ed.REG_MOTOR_DISTANCE_16)

    # Convert to the correct units - inches
    which /= 20    # not 20.3, but this is close enough

    return which


def Ed_AndModuleRegister8Bit(mod, reg, value):
    temp = Ed_ReadModuleRegister8Bit(mod, reg)
    Ed_WriteModuleRegister8Bit(mod, reg, (temp & value))


# -------------------------------------------------------------

# FUNCTIONS IMPLEMENTED IN COMPILER
# Listed here so that can show the compiler whether the function returns a
# value or not

#
# Any updates to these functions need to be changed in:
# edpy_values.signatures, edpy_code.CODE,
# compiler.SPECIALLY_HANDLED_FUNCTIONS, compiler.HandleSpecialCall
#


# PYTHON FUNCTIONS

def ord(character):
    # IMPLEMENTED IN COMPILER
    return 0

def chr(number):
    # IMPLEMENTED IN COMPILER
    return 0


def len(array):
    # IMPLEMENTED IN COMPILER
    return 0



# SIMPLE MOTOR FUNCTIONS IMPLEMENTED IN COMPILER

def Ed_SimpleDriveForwardRight():
    # IMPLEMENTED IN COMPILER
    pass


def Ed_SimpleDriveForwardLeft():
    # IMPLEMENTED IN COMPILER
    pass


def Ed_SimpleDriveStop():
    # IMPLEMENTED IN COMPILER
    pass


def Ed_SimpleDriveForward():
    # IMPLEMENTED IN COMPILER
    pass


def Ed_SimpleDriveBackward():
    # IMPLEMENTED IN COMPILER
    pass


def Ed_SimpleDriveBackwardRight():
    # IMPLEMENTED IN COMPILER
    pass


def Ed_SimpleDriveBackwardLeft():
    # IMPLEMENTED IN COMPILER
    pass


def Ed_Drive_INLINE_UNLIMITED(a, b, c):
    # IMPLEMENTED IN COMPILER
    pass


def Ed_DriveLeftMotor_INLINE_UNLIMITED(a, b, c):
    # IMPLEMENTED IN COMPILER
    pass


def Ed_DriveRightMotor_INLINE_UNLIMITED(a, b, c):
    # IMPLEMENTED IN COMPILER
    pass


# INTERNAL FUNCTIONS

def Ed_List1(size):
    # IMPLEMENTED IN COMPILER
    return 0


def Ed_List2(size, initial):
    # IMPLEMENTED IN COMPILER
    return 0


def Ed_TuneString1(size):
    # IMPLEMENTED IN COMPILER
    return 0


def Ed_TuneString2(size, initial):
    # IMPLEMENTED IN COMPILER
    return 0


def Ed_CreateObject(name):
    # IMPLEMENTED IN COMPILER
    pass


def Ed_RegisterEventHandler(event, function):
    # IMPLEMENTED IN COMPILER
    pass


def Ed_WriteModuleRegister8Bit(mod, reg, value):
    # IMPLEMENTED IN COMPILER
    pass


def Ed_WriteModuleRegister16Bit(mod, reg, value):
    # IMPLEMENTED IN COMPILER
    pass


def Ed_ClearModuleRegisterBit(mod, reg, bit):
    # IMPLEMENTED IN COMPILER
    pass


def Ed_SetModuleRegisterBit(mod, reg, bit):
    # IMPLEMENTED IN COMPILER
    pass


def Ed_ReadModuleRegister8Bit(mod, reg):
    # IMPLEMENTED IN COMPILER
    return 0


def Ed_ReadModuleRegister16Bit(mod, reg):
    # IMPLEMENTED IN COMPILER
    return 0


def Ed_ObjectAddr(obj):
    # IMPLEMENTED IN COMPILER
    return 0

"""



#!/usr/bin/env python


# * **************************************************************** **
#
# File: token_tool.py
# Desc: Assembler and dissassembler of tokens
# Note:
#
# Author: Brian Danilko, Likeable Software (brian@likeablesoftware.com)
#
# Copyright 2006, 2014, 2016, 2107 Microbric Pty Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License (in the doc/licenses directory)
# for more details.
#
# * **************************************************************** */

from __future__ import print_function
from __future__ import absolute_import

from . import io

import os
import os.path

from . import tokens
from . import hl_parser
from . import program


ROW_LENGTH = 14
COL_LENGTH = 6

# File scope objects
token_stream = None
err = None


def reset_tokens():
    global token_stream
    token_stream = None


def AsmError_NO_RET(number, internalError=None, line=0):
    if (internalError):
        io.Out.ErrorRaw(internalError)

    io.Out.Error(io.TS.ASM_INTERNAL_ERROR, "file::: Assembler internal error {0}", number)
    raise program.AssemblerError


def assem_file(f_name, called_from=[], pass_err=None):
    io.Out.DebugRaw("Assem_file:%s" % (f_name))

    io.Out.SetErrorRawContext(2, "reading token file: %s" % (f_name))

    # open file
    if (not os.path.isfile(f_name) or not os.access(f_name, os.R_OK)):
        print("ERROR - file %s doesn't exist or isn't readable" % (f_name))
        if (called_from):
            print("  Insert history: ", end='')
            output = ""
            for f, l in called_from:
                if (output):
                    output += ", "
                output += "%s:%d" % (f, l)
            print(output)
        return False

    fh = file(f_name, 'rU')

    # get each line but insert a file if find 'INSERT TOKENS'
    line_num = 1
    for line in fh:
        # if 'INPUT TOKENS fname' push existing file
        insert_check = line.split()
        if ((len(insert_check) >= 3) and
            (insert_check[0] == 'INSERT') and
            (insert_check[1].lower() == 'tokens')):
            called_from.append((f_name, line_num))
            assem_file(insert_check[2], called_from)
            called_from.pop()
        else:
            assem_line(line)

        line_num += 1

    return True


def assem_lines(lines):
    io.Out.DebugRaw("Assem_lines - lines:" + str(len(lines)))

    io.Out.SetErrorRawContext(2, "Assem_lines - lines:" + str(len(lines)))

    # get each line but insert a file if find 'INSERT TOKENS'
    line_num = 1
    for line in lines:
        assem_line(line)
        line_num += 1

    return True


def assem_line(line):
    """First scan of the operator to see which function to call"""
    io.Out.SetErrorRawContext(3, line)

    words = hl_parser.chop_line(line)

    if (not words):
        return

    t1 = words[0]

    io.Out.SetErrorRawContext(4, hl_parser.format_word_list(words))

    if (t1.type() == "label"):
        assem_spec_label(t1, words[1:], line=line)

    elif (t1.type() == "op"):
        op = t1.val()
        words = words[1:]

        if (op.endswith('b') or op.endswith('w')):
            size = 0
            if (op.endswith('w')):
                size = 1

            op = op[:-1]
            if (op == "mov"):
                assem_move(size, words, "", line=line)
            elif (op == "mol"):
                assem_move(size, words, "mol", line=line)
            elif (op == "dat"):
                assem_data(size, words, line=line)
            elif (op in ["not", "dec", "inc"]):
                assem_uni_math(op, size, words, line=line)
            elif (op in ["add", "sub", "mul", "cmp"]):
                assem_basic_math(op, size, words, line=line)
            elif (op in ["shl", "shr", "div", "mod"]):
                assem_other_math(op, size, words, line=line)
            elif (op in ["push", "pop", "stra", "stwa"]):
                assem_stack(op, size, words, line=line)
            elif (op in ["or", "and", "xor"]):
                assem_other_math(op, size, words, line=line)
            elif (op == "out"):
                assem_debug_output(op, size, words, line=line)
            else:
                AsmError_NO_RET(1, "Unknown operator:%s" % (op))
        else:
            if (op == "mova"):
                assem_move(0, words, "lcd", line=line)
            elif (op == "movtime"):
                assem_move(1, words, "time", line=line)
            elif (op in ["conv", "convl", "convm", "cmptime"]):
                # conv 8->16, conv 16->8 lsb, conv 16->8 msb, cmp time
                assem_conv(op, words, line=line)
            elif (op in ["disable", "enable", "error"]):
                assem_event(op, words, line=line)
            elif (op in ["ret", "dbnz", "dsnz"]):
                assem_jump(op, "", words, line=line)
            elif (op in ["bra", "bre", "brne", "brgr", "brge", "brl", "brle", "brz", "brnz"]):
                assem_jump("branch", op[2:], words, line=line)
            elif (op in ["suba", "sube", "subne", "subgr", "subge", "subl", "suble", "subz", "subnz"]):
                assem_jump("sub", op[3:], words, line=line)
            elif (op in ["stop", "bitset", "bitclr"]):
                assem_misc(op, words, line=line)
            elif (op in ["or", "and", "xor"]):
                # if there is no size after the op (which is now depricated) then treat it like a 'b'
                assem_other_math(op, 0, words, line=line)
            elif (op in ["stinc", "stdec", "push"]):
                assem_stack_math(op, words, line=line)

            elif (op in ["DATB", "DATW"]):
                assem_spec_data(op[3], words, line=line)
            elif (op == "DATA"):
                assem_spec_data_lcd(op[3], words, line=line)
            elif (op in ["BINB"]):
                assem_spec_binary(op[3], words, line=line)
            elif (len(op) == 7 and op[:6] == "RESERV"):
                assem_spec_reserve(op[6], words, line=line)
            elif (op in ["BEGIN", "END"]):
                assem_spec_begin_end(op, words, line=line)
            elif (op == "VERSION"):
                assem_spec_version(words, line=line)
            elif (op == "LIMITS"):
                assem_spec_limits(words, line=line)
            elif (op == "DEVICE"):
                assem_spec_device(words, line=line)
            elif (op == "INSERT"):
                assem_spec_insert(words, line=line)
#           elif (op == "COMMS"):
#               assem_spec_comms(words, line=line)
            elif (op == "FINISH"):
                assem_spec_finish(words, line=line)
            else:
                AsmError_NO_RET(2, "Unknown operator:%s" % (op))


def assem_move(size, words, special, line):
    io.Out.DebugRaw("Move size:%d, words:%s" % (size, hl_parser.format_word_list(words)))
    if (len(words) != 2):
        AsmError_NO_RET(3, "Move needs 2 arguments")

    # check the special case of move from the accumulator
    elif ((special == "mol") or
          ((special == "") and (words[0].type() == "modreg") and (words[0].val() == 0xf0))):
        assem_move_from_acc(size, words, special, line)
    else:
        assem_move_not_from_acc(size, words, special, line)


def assem_move_from_acc(size, words, special, line):

    token = tokens.Token("misc", err, line)

    if (special == "mol"):
        if (size != 0):
            AsmError_NO_RET(4, "mol only has a byte variant (not word)")
        elif ((words[0].type() != "var") or
              (words[1].type() != "modreg")):
            AsmError_NO_RET(5, "Invalid arguments for molb")
        else:
            token.add_byte(0, 0x34)                     # op code
            token.add_byte(1, 0)                        # 16-bit var offset
            token.add_vname(1, 1, words[0].val())       # second 1 means 16-bit
            token.add_byte(2, words[1].val())           # modreg code

    else:
        token.add_bits(0, 4, 0xf, 0x3)  # top 4 bits = 0011

        # destination
        if (words[1].type() == "modreg"):
            if (size == 0):
                token.add_bits(0, 0, 0xf, 0x2)
            else:
                token.add_bits(0, 0, 0xf, 0x3)
            token.add_byte(1, words[1].val())

        elif (words[1].type() in ["var", "arg"]):
            if (size == 0):
                token.add_bits(0, 0, 0xf, 0x0)
            else:
                token.add_bits(0, 0, 0xf, 0x1)

            if (words[1].type() == "arg"):
                token.add_byte(1, words[1].num())
            else:
                token.add_byte(1, 0)
                token.add_vname(1, size, words[1].val())

        else:
            AsmError_NO_RET(6, "Destination must be a mod/reg or variable")

    token.finish(token_stream)


def assem_move_not_from_acc(size, words, special, line):

    token = tokens.Token("move", err, line)
    token.add_bits(0, 6, 3, 1)
    sdindex = 1

    # source
    if (words[0].type() == "const"):
        if (not special and words[0].val() >= 0 and words[0].val() <= 3):
            token.add_bits(0, 3, 0x3, words[0].val())
            token.add_bits(0, 2, 0x1, size)
        else:
            if (size == 0):
                token.add_bits(0, 2, 0xf, 0xa)
                token.add_byte(sdindex, words[0].val())
                sdindex += 1
            else:
                if (not special):
                    token.add_bits(0, 2, 0xf, 0xb)
                elif (special == "time"):
                    token.add_bits(0, 2, 0xf, 0xf)
                else:
                    AsmError_NO_RET(7, "Source for lcd moves must be byte sized")

                token.add_word(sdindex, words[0].val())
                sdindex += 2
    elif (words[0].type() == "modreg"):
        if (size == 0):
            token.add_bits(0, 2, 0xf, 0x8)
        else:
            token.add_bits(0, 2, 0xf, 0x9)
        token.add_byte(sdindex, words[0].val())
        sdindex += 1

    elif (words[0].type() in ["var", "arg"]):
        if (size == 0):
            token.add_bits(0, 2, 0xf, 0xc)
        else:
            if (not special):
                token.add_bits(0, 2, 0xf, 0xd)
            elif (special == "time"):
                token.add_bits(0, 2, 0xf, 0xe)
            else:
                AsmError_NO_RET(8, "Source for lcd moves must be byte sized")

        if (words[0].type() == "arg"):
            token.add_byte(sdindex, words[0].num())
        else:
            token.add_byte(sdindex, 0)
            token.add_vname(sdindex, size, words[0].val())
        sdindex += 1

    # if this is a time chkpt then the destination must be in 8bit space
    if (special == "time"):
        size = 0

    # destination
    if (words[1].type() == "modreg"):
        if (special):
            AsmError_NO_RET(9, "Destination for lcd and time moves can NOT be mod/reg")
        # ((words[0].type() == "modreg") and (words[0].val() == 0xf0))):

        # special processing for move to ACC
        if (words[1].val() == 0xf0) :
            token.add_bits(0, 0, 0x3, 0x3)
        else:
            token.add_bits(0, 0, 0x3, 0x0)
            token.add_byte(sdindex, words[1].val())
            sdindex += 1

    elif (words[1].type() in ["var", "arg"]):
        if (size == 0):
            if (special != "lcd"):
                token.add_bits(0, 0, 0x3, 0x1)
            else:
                token.add_bits(0, 0, 0x3, 0x3)
        else:
            if (special):
                AsmError_NO_RET(10, "Desination for lcd and time moves can not be in word variables")

            token.add_bits(0, 0, 0x3, 0x2)

        if (words[1].type() == "arg"):
            token.add_byte(sdindex, words[1].num())
        else:
            token.add_byte(sdindex, 0)
            if (special == "lcd"):
                token.add_vname(sdindex, 2, words[1].val())
            else:
                token.add_vname(sdindex, size, words[1].val())
        sdindex += 1

    io.Out.DebugDumpObjectRaw(token, "move")
    token.finish(token_stream)


def assem_data(size, words, line):
    io.Out.DebugRaw("data size:%s words:%s" % (size, hl_parser.format_word_list(words)))

    if (len(words) < 3):
        AsmError_NO_RET(11, "dat[bw] needs at least 3 arguments: var, len and value1")

    if (words[0].type() not in ["arg", "var"]):
        AsmError_NO_RET(12, "dat[bw] first argument must be a variable or location")
    elif (words[0].type() == "arg"):
        start = words[0].anum()
        name = '*'
    else:
        start = 0
        name = words[0].val()

    length = 1
    if (words[1].val() == '*'):
        length = -1
    else:
        length = words[1].anum()

    words = words[2:]
    values = []
    # now words are the values
    for w in words:
        if (w.type() not in ["arg", "string"]):
            AsmError_NO_RET(13, "Word should have been an argument or string!")
        else:
            if (w.type() == "string"):
                for c in w.val():
                    values.append(ord(c))
            else:
                values.append(w.anum())

    # Do space sanity checks
    if (length > 0 and len(values) > length):
        AsmError_NO_RET(14, "Data has length of %d but more values (%d) then length" % \
                                   (length, len(values)))

    # zero up to the named length
    if (length > len(values)):
        while (len(values) < length):
            values.append(0)

    # Make the tokens
    if (len(values)):

        tokens_to_create = (len(values) + 14) / 15
        last = len(values) % 15

        # print tokens_to_create, last
        val_index = 0
        for i in range(tokens_to_create):
            token = tokens.Token("data", err, line)
            token.add_bits(0, 4, 3, size + 1)
            if ((i < (tokens_to_create - 1)) or (last == 0)):
                val_count = 15
            else:
                val_count = last

            token.add_bits(0, 0, 0xf, val_count)

            # add the location
            if (name != '*'):
                token.add_byte(1, i * 15)
                token.add_vname(1, size, name)
            else:
                token.add_byte(1, start + (i * 15))

            token_index = 2
            # add the data
            if (size == 0):
                for j in range(val_count):
                    # print(val_index)
                    token.add_byte(token_index, values[val_index])
                    token_index += 1
                    val_index += 1
            else:
                for j in range(val_count):
                    # print(val_index)
                    token.add_word(token_index, values[val_index])
                    token_index += 2
                    val_index += 1

            token.finish(token_stream)


uni_assem_map = {"not": 0, "inc": 1, "dec": 2}


def assem_uni_math(op, size, words, line):
    io.Out.DebugRaw("Mathu %s size:%d, words:%s" % (op, size, hl_parser.format_word_list(words)))
    if (len(words) > 1):
        AsmError_NO_RET(15, "Unary Math has at most one argument")

    token = tokens.Token("uni-math", err, line)
    token.add_bits(0, 6, 3, 2)
    token.add_bits(0, 3, 1, size)
    token.add_bits(0, 0, 3, uni_assem_map[op])

    # if no argument or %_cpu:acc, then this is using the ACC
    if (not words or
        (words[0].type() == "modreg" and words[0].val() == 0xf0)):
        token.add_bits(0, 2, 1, 0)
        token.finish(token_stream)
    elif (words[0].type() in ["var", "arg"]):
        token.add_bits(0, 2, 1, 1)
        if (words[0].type() == "arg"):
            token.add_byte(1, words[0].num())
        else:
            token.add_byte(1, 0)
            token.add_vname(1, size, words[0].val())
        token.finish(token_stream)
    else:
        AsmError_NO_RET(16, "Unary Math - invalid argument type")

basic_assem_map = {"add": 0, "sub": 1, "mul": 2, "cmp": 3}


def assem_basic_math(op, size, words, line):
    io.Out.DebugRaw("Mathb %s size:%s, words:%s" % (op, size, hl_parser.format_word_list(words)))
    if (len(words) != 1):
        AsmError_NO_RET(17, "Basic Math needs one argument")

    token = tokens.Token("basic-math", err, line)
    token.add_bits(0, 6, 3, 2)
    token.add_bits(0, 4, 3, 1)
    token.add_bits(0, 3, 1, size)
    token.add_bits(0, 0, 3, basic_assem_map[op])

    if (words[0].type() in ["var", "arg"]):
        # math with a variable
        token.add_bits(0, 2, 1, 0)
        if (words[0].type() == "arg"):
            token.add_byte(1, words[0].num())
        else:
            token.add_byte(1, 0)
            token.add_vname(1, size, words[0].val())
        token.finish(token_stream)
    elif (words[0].type() == "const"):
        token.add_bits(0, 2, 1, 1)
        if (size == 0):
            token.add_byte(1, words[0].val())
        else:
            token.add_word(1, words[0].val())

        token.finish(token_stream)
    else:
        AsmError_NO_RET(18, "Basic Math - invalid argument type")

other1_assem_map = {"shl": 0, "shr": 1, "div": 2, "mod": 3}
other2_assem_map = {"or": 0, "and": 1, "xor": 2}


def assem_other_math(op, size, words, line):
    io.Out.DebugRaw("Matho %s size:%d, words:%s" % (op, size, hl_parser.format_word_list(words)))

    if (len(words) != 1):
        AsmError_NO_RET(19, "Logic Math needs one arguement")

    # if ((op in ("shl", "shr")) and (size == 1) and
    #     (words[0].type() == "const")):
    #     AsmError_NO_RET(20, "Shifts by constants only take 8-bit constants")

    token = tokens.Token("log-math", err, line)
    token.add_bits(0, 6, 3, 2)
    token.add_bits(0, 3, 1, size)

    if (op in other1_assem_map.keys()):
        token.add_bits(0, 4, 3, 2)
        token.add_bits(0, 0, 3, other1_assem_map[op])
    else:
        token.add_bits(0, 4, 3, 3)
        token.add_bits(0, 0, 3, other2_assem_map[op])

    if (words[0].type() in ["var", "arg"]):
        # math with a variable
        token.add_bits(0, 2, 1, 0)
        if (words[0].type() == "arg"):
            token.add_byte(1, words[0].val())
        else:
            token.add_byte(1, 0)
            token.add_vname(1, size, words[0].val())
        token.finish(token_stream)
    elif (words[0].type() == "const"):
        token.add_bits(0, 2, 1, 1)
        if ((size == 0) or (op in ("shl", "shr"))):
            # shl/shr only take one byte since it's value is at most 0-15
            token.add_byte(1, words[0].val())
        else:
            token.add_word(1, words[0].val())

        token.finish(token_stream)
    else:
        AsmError_NO_RET(21, "Logic Math - invalid argument type")


def assem_conv(op, words, line):
    io.Out.DebugRaw("Conv %s words:%s" % (op, hl_parser.format_word_list(words)))

    token = tokens.Token("conv", err, line)

    if (op == "cmptime"):
        if (len(words) != 1):
            AsmError_NO_RET(22, "Cmptime needs one arguement")
        else:
            token.add_bits(0, 6, 3, 2)
            token.add_bits(0, 3, 1, 0)
            token.add_bits(0, 0, 7, 7)
            if (words[0].type() == "arg"):
                token.add_byte(1, words[0].num())
            elif (words[0].type() == "var"):
                token.add_byte(1, 0)
                token.add_vname(1, 0, words[0].val())
            else:
                AsmError_NO_RET(23, "Cmptime takes a variable as it's argument")
    else:
        if (len(words) != 0):
            AsmError_NO_RET(24, "Conversions don't that arguments")
        else:
            token.add_bits(0, 6, 3, 2)
            token.add_bits(0, 0, 7, 3)
            if (op == "convm"):
                token.add_bits(0, 3, 1, 1)

    token.finish(token_stream)


def assem_stack(op, size, words, line):
    io.Out.DebugRaw("Stack %s size:%d, words:%s" % (op, size, hl_parser.format_word_list(words)))
    # op in ['push', 'pop', 'stra', 'stwa'], size in [0, 1], one word
    if (len(words) != 1):
        AsmError_NO_RET(25, "Stack ops need 1 argument")

    token = tokens.Token("stack", err, line)
    token.add_bits(0, 6, 3, 3)
    token.add_bits(0, 5, 1, 1)
    token.add_bits(0, 4, 1, size)

    if (op == "push"):
        if (words[0].type() == 'const'):
            if (size == 0):
                token.add_byte(1, words[0].val())
            else:
                token.add_word(1, words[0].val())
        elif (words[0].type() == 'modreg'):
            # don't do special ACC here, as the size [b,w] was used.
            token.add_bits(0, 0, 0xf, 2)
            token.add_byte(1, words[0].val())
        elif (words[0].type() in ["var", "arg"]):
            token.add_bits(0, 0, 0xf, 1)
            if (words[0].type() == "arg"):
                token.add_byte(1, words[0].num())
            else:
                token.add_byte(1, 0)
                token.add_vname(1, size, words[0].val())
        else:
            AsmError_NO_RET(26, "Push - invalid operand: %s" % (words[0].type()))

    elif (op == "pop"):
        if (words[0].type() == 'modreg'):
            if (words[0].val() == 0xf0):  # ACC
                token.add_bits(0, 0, 0xf, 0x4)
            else:
                token.add_bits(0, 0, 0xf, 6)
                token.add_byte(1, words[0].val())
        elif (words[0].type() in ["var", "arg"]):
            token.add_bits(0, 0, 0xf, 5)
            if (words[0].type() == "arg"):
                token.add_byte(1, words[0].num())
            else:
                token.add_byte(1, 0)
                token.add_vname(1, size, words[0].val())
        else:
            AsmError_NO_RET(27, "Pop - invalid operand: %s" % (words[0].type()))

    elif (op == "stra"):  # read from stack into acc
        token.add_bits(0, 0, 0xf, 9)
        if (words[0].type() == "const"):
            token.add_byte(1, words[0].val())
        elif (words[0].type() == "arg"):
            token.add_byte(1, words[0].num())
        else:
            AsmError_NO_RET(28, "Stra - invalid operand: %s" % (words[0].type()))

    else:  # (op == "stwa"):  # write from acc into stack
        token.add_bits(0, 0, 0xf, 0xc)
        if (words[0].type() == "const"):
            token.add_byte(1, words[0].val())
        elif (words[0].type() == "arg"):
            token.add_byte(1, words[0].num())
        else:
            AsmError_NO_RET(29, "Strw - invalid operand: %s" % (words[0].type()))

    token.finish(token_stream)


def assem_debug_output(op, size, words, line):
    io.Out.DebugRaw("Debug output %s size:%d, words:%s" % (op, size, hl_parser.format_word_list(words)))
    # op in ['out'], size in [0, 1], one word
    if (len(words) != 1):
        AsmError_NO_RET(30, "Debug output needs 1 argument")

    if (words[0].type() != "var"):
        AsmError_NO_RET(31, "Error - Debug output only handles variables - invalid operand: %s" % (words[0].type()))

    token = tokens.Token("output", err, line)
    token.add_bits(0, 6, 3, 3)
    token.add_bits(0, 5, 1, 1)
    token.add_bits(0, 4, 1, size)
    token.add_bits(0, 0, 0x0f, 0x0e)

    # this is a variable - checked above
    token.add_byte(1, 0)
    token.add_vname(1, size, words[0].val())

    token.finish(token_stream)


def assem_stack_math(op, words, line):
    io.Out.DebugRaw("Stack math %s words:%s" % (op, hl_parser.format_word_list(words)))
    # op in ['stinc', 'stdec', 'push'], one word
    if (len(words) != 1):
        AsmError_NO_RET(32, "Stack ops need 1 argument")

    token = tokens.Token("stack", err, line)
    token.add_bits(0, 4, 0xf, 0xe)    # top 3 bits set

    if (op == "push"):
        if (words[0].type() == 'modreg'):
            if (words[0].val() == 0xf0):  # ACC
                token.add_bits(0, 0, 0xf, 0x3)
            else:
                AsmError_NO_RET(33, "Push with no size - invalid operand: %s" % (words[0].type()))
        else:
            AsmError_NO_RET(34, "Push with no size - invalid operand: %s" % (words[0].type()))

    elif (op == "stinc"):
        token.add_bits(0, 0, 0xf, 0xa)
        if (words[0].type() == "const"):
            token.add_byte(1, words[0].val())
        elif (words[0].type() == "arg"):
            token.add_byte(1, words[0].num())
        else:
            AsmError_NO_RET(35, "Stinc - invalid operand: %s" % (words[0].type()))
    else:  # stdec
        token.add_bits(0, 0, 0xf, 0xb)
        if (words[0].type() == "const"):
            token.add_byte(1, words[0].val())
        elif (words[0].type() == "arg"):
            token.add_byte(1, words[0].num())
        else:
            AsmError_NO_RET(36, "Stdec - invalid operand: %s" % (words[0].type()))

    token.finish(token_stream)


def assem_event(op, words, line):
    io.Out.DebugRaw("Event %s words:%s" % (op, hl_parser.format_word_list(words)))

    token = tokens.Token("event", err, line)

    if (op == "error"):
        if (len(words) != 1):
            AsmError_NO_RET(37, "Error needs 1 argument")
        token.add_byte(0, 0xef)

        if (words[0].type() == "const"):
            token.add_byte(1, words[0].val())
        elif (words[0].type() == "arg"):
            token.add_byte(1, words[0].num())
        else:
            AsmError_NO_RET(38, "Error - invalid operand: %s" % (words[0].type()))

    else:
        # op is one of: enable, disable, no words
        if (len(words) != 0):
            AsmError_NO_RET(39, "Enable/disable don't take arguments")

        token.add_bits(0, 6, 3, 2)
        token.add_bits(0, 4, 3, 3)
        token.add_bits(0, 0, 3, 3)

        if (op == 'enable'):
            token.add_bits(0, 2, 1, 1)
        else:  # (op == 'disable')
            token.add_bits(0, 2, 1, 0)

    token.finish(token_stream)


jcond_assem_map = {'a': 0, 'e': 1, 'ne': 2, 'gr': 3, 'ge': 4, 'l': 5, 'le': 6, 'z': 1, 'nz': 2}


def assem_jump(op, cond, words, line):
    io.Out.DebugRaw("Jump %s cond:%s, words:%s" % (op, cond, hl_parser.format_word_list(words)))
    # ops is one of: branch, sub, ret, dbnz, dsnz
    # cond is one of: a, e, ne, g, l, le, lg or empty for ret, d?nz
    token = tokens.Token("jump", err, line)
    token.add_bits(0, 6, 3, 3)

    if (op == "ret"):
        if (len(words) != 0):
            AsmError_NO_RET(40, "Ret don't take an argument")
        else:
            token.add_bits(0, 0, 0x3f, 0x28)
    elif (len(words) != 1):
        AsmError_NO_RET(41, "Jumps need a target to jump to")
    else:
        if (op[:2] in ['su', 'ds']):
            # a call to a subroutine - push a frame
            token.add_bits(0, 3, 1, 1)
        else:
            token.add_bits(0, 3, 1, 0)

        if (not cond):
            # one of dbnz or dsnz
            token.add_bits(0, 0, 7, 7)
        else:
            token.add_bits(0, 0, 7, jcond_assem_map[cond])

        # now the target
        if (words[0].type() == "const"):
            offset = words[0].val()
            if (offset >= tokens.MIN_SBYTE and offset <= tokens.MAX_SBYTE):
                if (offset < 0):
                    # convert to a signed offset
                    offset += 256
                token.add_byte(1, offset)
            elif (offset >= tokens.MIN_WORD and offsec <= tokens.MAX_WORD):
                token.add_bits(0, 4, 1, 1)
                token.add_word(1, offset)

        elif (words[0].type() == "label"):
            if (words[0].val().startswith(':')):
                # globals are always long jumps
                token.add_bits(0, 4, 1, 1)
                token.set_jump_label(1, words[0].val(), True)
                token.add_word(1, 0)    # A placeholder
            else:
                token.set_jump_label(1, words[0].val())
                token.add_byte(1, 0)    # A placeholder

        else:
            AsmError_NO_RET(42, "Jumps need either a constant or a label as argument, not a: " + words[0].type())

    token.finish(token_stream)


def assem_misc(op, words, line):
    io.Out.DebugRaw("Misc op:%s words:%s" % (op, hl_parser.format_word_list(words)))
    if (op == "stop"):
        if (len(words) != 0):
            AsmError_NO_RET(43, "Stop doesn't take arguments")

        token = tokens.Token("misc", err, line)
        token.add_bits(0, 0, 0xff, 0xff)
        token.finish(token_stream)

    elif (op in ["bitset", "bitclr"]):
        if (len(words) != 2):
            AsmError_NO_RET(44, "Bitset/bitclr needs 2 arguments: bit and mod/reg")

        bit = words[0].anum()
        if (bit < 0 or bit > 7):
            AsmError_NO_RET(45, "Bitset/bitclr bit must be between 0 and 7 (not %d)" % (bit))

        modreg = words[1].amodreg()

        token = tokens.Token("misc", err, line)
        token.add_bits(0, 4, 0x0f, 0x00)
        if (op == "bitset"):
            token.add_bits(0, 3, 0x1, 0x1)
        else:
            token.add_bits(0, 3, 0x1, 0x0)

        token.add_bits(0, 0, 0x7, bit)
        token.add_byte(1, modreg)
        token.finish(token_stream)

    else:
        AsmError_NO_RET(46, "Unknown misc operator: " + op)

# ********* SPECIALS (in caps) ********************************************


def assem_spec_label(label, words, line):
    io.Out.DebugRaw("Spec_label:%s" % (label.val()))
    if (len(words) != 0):
        AsmError_NO_RET(47, "A label doesn't take any arguments")

    token_stream.add_label(label.val())


def assem_spec_data(which, words, line):
    io.Out.DebugRaw("Spec_data which:%s words:%s" % (which, hl_parser.format_word_list(words)))

    if (len(words) < 2):
        AsmError_NO_RET(48, "DAT[BW] needs at least 2 arguments: name, start")

    space = tokens.space_types[which]

    if (words[0].val() == '*'):
        name = '*'
    else:
        name = words[0].astr()

    if (words[1].val() == '*'):
        start = -1
        if (name == '*'):
            AsmError_NO_RET(49, "DAT[BW] both name and loc can't both be '*'")
    else:
        start = words[1].anum()

    words = words[2:]

    length = 1
    if (len(words) >= 1):
        if (words[0].val() == '*'):
            length = -1
        else:
            length = words[0].anum()
        words = words[1:]

    values = []
    # now words are the values
    for w in words:
        if (w.type() not in ["arg", "string"]):
            AsmError_NO_RET(50, "Word should have been an argument or string!")
        else:
            if (w.type() == "string"):
                for c in w.val():
                    values.append(ord(c))
            else:
                values.append(w.anum())

    # Do space sanity checks
    if (length > 0 and len(values) > length):
        AsmError_NO_RET(51, "Data has length of %d but more values (%d) then length" %
                         (length, len(values)))

    if (length > 0):
        real_length = length
    else:
        real_length = len(values)

    if (start > 0):
        test = start + real_length
    else:
        # don't know where it will start so can't know if it will fit yet
        test = real_length

    if (test <= 0):
        AsmError_NO_RET(52, "There is no data for the space")
    elif (test >= 255):
        io.Out.Error(io.TS.ASM_MEM_OVERFLOW,
                     "file::: Overflowed {0} memory", which)
        raise program.AssemblerError

    # special case where we have an UNNAMED area with a length - zero beyond the values up to length
    if ((name == '*') and (length > 0) and (length > len(values))):
        while (len(values) < length):
            values.append(0)
        # print(values)

    # Make the tokens
    if (len(values)):

        tokens_to_create = (len(values) + 14) // 15
        last = len(values) % 15

        # print(tokens_to_create, last)
        val_index = 0
        for i in range(tokens_to_create):
            token = tokens.Token("data", err, line)
            token.add_bits(0, 4, 3, space + 1)
            if ((i < (tokens_to_create - 1)) or (last == 0)):
                val_count = 15
            else:
                val_count = last

            token.add_bits(0, 0, 0xf, val_count)

            # add the location
            if (name != '*'):
                token.add_byte(1, i * 15)
                token.add_vname(1, space, name)
            else:
                token.add_byte(1, start + (i * 15))

            token_index = 2
            # add the data
            if (which == "B"):
                for j in range(val_count):
                    # print(val_index)
                    token.add_byte(token_index, values[val_index])
                    token_index += 1
                    val_index += 1
            else:
                for j in range(val_count):
                    # print(val_index)
                    if (values[val_index] > tokens.MAX_WORD):
                        token.add_uword(token_index, values[val_index])
                    else:
                        token.add_word(token_index, values[val_index])
                    token_index += 2
                    val_index += 1

            token.finish(token_stream)

    # Add the info about the variable to the token_stream
    if (name != '*'):
        token_stream.add_variable(space, name, start, real_length)


def assem_spec_data_lcd(which, words, line):
    io.Out.DebugRaw("Spec_data_lcd which:%s words:%s" % (which, hl_parser.format_word_list(words)))

    if (len(words) < 4):
        AsmError_NO_RET(53, "DATA needs at least 4 arguments: row, col, len, val1")

    if (words[0].val() == '*'):
        AsmError_NO_RET(54, "DATA needs a real row")
    else:
        row = words[0].anum()

    if (words[1].val() == '*'):
        AsmError_NO_RET(55, "DATA needs a real column")
    else:
        column = words[1].anum()

    start = row * ROW_LENGTH + column

    if (words[2].val() == '*'):
        length = -1
    else:
        length = words[2].anum()

    words = words[3:]

    values = []
    # now words are the values
    for w in words:
        if (w.type() not in ["arg", "string"]):
            AsmError_NO_RET(56, "Word should have been an argument or string!")
        else:
            if (w.type() == "string"):
                for c in w.val():
                    values.append(ord(c))
            else:
                values.append(w.anum())

    # Do space sanity checks
    if (length > 0 and len(values) > length):
        AsmError_NO_RET(57, "Data has length of %d but more values (%d) then length" %
                         (length, len(values)))

    if (length > 0):
        real_length = length
    else:
        real_length = len(values)

    if (start + real_length > ROW_LENGTH * COL_LENGTH):
        io.Out.Error(io.TS.ASM_MEM_OVERFLOW,
                     "file::: Overflowed {0} memory", which)
        raise program.AssemblerError

    # Add spaces if needed
    if ((length > 0) and (length > len(values))):
        while (len(values) < length):
            values.append(' ')

    # Make the tokens
    if (len(values)):

        tokens_to_create = (len(values) + 14) / 15
        last = len(values) % 15

        # print(tokens_to_create, last)
        val_index = 0
        for i in range(tokens_to_create):
            token = tokens.Token("data", err, line)
            token.add_bits(0, 4, 3, 3)
            if ((i < (tokens_to_create - 1)) or (last == 0)):
                val_count = 15
            else:
                val_count = last

            token.add_bits(0, 0, 0xf, val_count)

            # add the row/column
            loc = start + i * 15
            token.add_byte(1, loc / ROW_LENGTH)
            token.add_byte(2, loc % ROW_LENGTH)

            token_index = 3
            # add the data
            for j in range(val_count):
                # print(val_index)
                token.add_byte(token_index, values[val_index])
                token_index += 1
                val_index += 1

            token.finish(token_stream)


def assem_spec_binary(which, words, line):
    io.Out.DebugRaw("Spec_binary which:%s words:%s" % (which, hl_parser.format_word_list(words)))

    token = tokens.Token("binary", err, line)
    token_index = 0

    # now words are the values
    for w in words:
        if (w.type() not in ["arg", "string", "const"]):
            AsmError_NO_RET(58, "Word should have been an argument or string!")
        else:

            if (w.type() == "string"):
                for c in w.val():
                    token.add_byte(token_index, ord(c))
                    token_index += 1
            else:
                token.add_byte(token_index, w.anum())
                token_index += 1

    token.finish(token_stream)


def assem_spec_reserve(which, words, line):
    io.Out.DebugRaw("Spec_reserve which:%s words:%s" % (which, hl_parser.format_word_list(words)))
    if (len(words) != 2):
        AsmError_NO_RET(59, "RESERV[ABW] needs 2 arguments: start, length")

    token_stream.reserve_name_space(tokens.space_types[which], words[0].anum(), words[1].anum())


def assem_spec_version(words, line):
    io.Out.DebugRaw("Spec_version words:%s" % (hl_parser.format_word_list(words)))
    if (len(words) != 2):
        AsmError_NO_RET(60, "VERSION needs 2 arguments: major, minor")
    major = words[0].anum()
    minor = words[1].anum()

    if (major < 0 or major > 15):
        AsmError_NO_RET(61, "major version must be between 0 and 15 (not %d)" % major)

    if (minor < 0 or minor > 15):
        AsmError_NO_RET(62, "minor version must be between 0 and 15 (not %d)" % major)

    token_stream.add_version(major, minor)


def assem_spec_begin_end(op, words, line):
    io.Out.DebugRaw("Spec_begin_end op:%s words:%s" % (op, hl_parser.format_word_list(words)))
    if (len(words) < 1):
        AsmError_NO_RET(63, "BEGIN/END need a type argument")

    if (words[0].astr() == "FIRMWARE"):
        if (op == "BEGIN"):
            if (len(words) != 1):
                AsmError_NO_RET(64, "FIRMWARE doesn't take any arguments.")
            token_stream.add_begin("firmware")
        else:
            # END MAIN
            token_stream.add_end("firmware")
    elif (words[0].astr() == "EVENT"):
        if (op == "BEGIN"):
            if (len(words) != 4):
                AsmError_NO_RET(65, "EVENT needs 3 arguments: mod/reg, mask, value")
            modreg = words[1].amodreg()
            mask = words[2].anum()
            value = words[3].anum()
            token_stream.add_begin("event", modreg, mask, value)
        else:
            # END EVENT
            token_stream.add_end("event")

    elif (words[0].astr() == "MAIN"):
        if (op == "BEGIN"):
            if (len(words) != 1):
                AsmError_NO_RET(66, "MAIN doesn't take any arguments.")
            token_stream.add_begin("main")
        else:
            # END MAIN
            token_stream.add_end("main")

    else:
        AsmError_NO_RET(67, "BEGIN/END needs one of: MAIN, EVENT, FIRMWARE")


def assem_spec_limits(words, line):
    io.Out.DebugRaw("Spec_limits words:%s" % (hl_parser.format_word_list(words)))
    if (len(words) != 5):
        AsmError_NO_RET(68, "LIMITS needs exactly 5 arguments")

    b_limit = words[0].anum()
    w_limit = words[1].anum()
    # In this token stream there is no LCD
    # l_limit = words[2].anum()
    l_limit = 0
    e_handlers = words[3].anum()
    t_bytes_limit = words[4].anum()

    token_stream.set_limits(b_limit, w_limit, l_limit, e_handlers, t_bytes_limit)


def assem_spec_device(words, line):
    io.Out.DebugRaw("Spec_device words:%s" % (hl_parser.format_word_list(words)))
    if (len(words) == 2):
        name = ""
    elif (len(words) == 3):
        name = words[2].astr()
    else:
        AsmError_NO_RET(69, "DEVICE needs 2 or 3 arguments")

    device_type = words[0].astr()
    location = words[1].anum()

    # error checking is done in the hl_parser
    # add the device to the high_level parser for parsing of subsequent names
    if (hl_parser.add_device(location, device_type, name)):
        token_stream.add_device(hl_parser.device_types[device_type], location,
                                hl_parser.device_storage[device_type])


def assem_spec_insert(words, line):
    io.Out.DebugRaw("Spec_insert words:%s" % (hl_parser.format_word_list(words)))
    if (len(words) != 2):
        AsmError_NO_RET(70, "INSERT needs type and filename arguments")

    type = words[0].astr()
    f_name = words[1].astr()

    if (type.lower() not in ['tokens', 'binary']):
        AsmError_NO_RET(71, "INSERT type must be one of: 'tokens', 'binary'")

    if (type.lower() == 'tokens'):
        AsmError_NO_RET(72, "INSERT TOKENS should have been consumed higher up! Eek!")

    # we have an insert binary special to deal with
    if (not os.path.isfile(f_name) or not os.access(f_name, os.R_OK)):
        AsmError_NO_RET(73, "INSERT BINARY file:%s doesn't exist or isn't readable")

    token = tokens.Token("binary", err, line)
    token.add_binary_file(f_name)
    token.finish(token_stream)


def assem_spec_comms(words, line):
    io.Out.DebugRaw("Spec_comms words:%s" % (hl_parser.format_word_list(words)))
    if (len(words) != 1):
        AsmError_NO_RET(74, "COMMS needs exactly 1 argument")

    token = tokens.Token("comms", err, line)
    token.add_bits(0, 0, 0xf, 0x8)

    token.add_uword(1, words[0].anum())
    token.finish(token_stream)

    token_stream.set_comms(words[0].anum())


def assem_spec_finish(words, line):
    io.Out.DebugRaw("Spec_finish words:%s" % (hl_parser.format_word_list(words)))
    if (len(words) != 0):
        AsmError_NO_RET(75, "FINISH doesn't have any arguments")

    token_stream.finish_tokens()


# *********** Sequencing the assembly *********************************

def assemble_file(srcPath, debug):
    global token_stream

    io.Out.Top(io.TS.ASM_START, "Starting assembler")
    ok = True

    try:
        # setup the token stream
        token_stream = tokens.TokenStream()
        token_stream.clear()
        ok = assem_file(srcPath, [])

    except program.EdPyError:
        ok = False
        if (io.Out.IsReRaiseSet()):
            raise

    except:
        ok = False
        io.Out.Error(io.TS.CMP_INTERNAL_ERROR,
                     "file::: Compiler internal error {0}", 703)
        if (io.Out.IsReRaiseSet()):
            raise

    if (not ok):
        # print("ERROR when assembling!")
        return [], "", "", (0, 0)

    # print (len(lines), lines[0])
    return finish_assembley(srcPath, debug)


def assemble_lines(lines, debug):
    global token_stream

    io.Out.Top(io.TS.ASM_START, "Starting assembler")
    ok = True

    try:
        # setup the token stream
        token_stream = tokens.TokenStream()
        token_stream.clear()
        ok = assem_lines(lines)

    except program.EdPyError:
        ok = False
        if (io.Out.IsReRaiseSet()):
            raise

    except:
        ok = False
        io.Out.Error(io.TS.CMP_INTERNAL_ERROR,
                     "file::: Compiler internal error {0}", 704)
        if (io.Out.IsReRaiseSet()):
            raise

    if (not ok):
        # print("ERROR when assembling!")
        return [], "", "", (0, 0)

    # print (len(lines), lines[0])
    return finish_assembley("internal", debug)


def finish_assembley(source, debug):
    ok = True
    try:
        token_analysis = tokens.TokenAnalyser(token_stream)
        token_analysis.map_all_variables()

        if (debug and not io.Out.WasErrorRaised()):
            hl_parser.dump_devices()
            token_analysis.dump_variable_map()

        token_analysis.fixup_jumps()

        if (debug and not io.Out.WasErrorRaised()):
            token_stream.dump_tokens(source)
            token_analysis.dump_extras()

        download_type, version, header, added_bytes = token_analysis.create_header()
        # print(download_type, version, header)

    except:
        ok = False
        if (io.Out.IsReRaiseSet()):
            raise

    if (not ok):
        # print("ERROR when assembling!")
        return [], "", "", (0, 0)

    # get the token bytes
    download_str = ""
    download_bytes = []
    for t in header:
        download_bytes.append(t)
        download_str += chr(t)

    for t in token_stream.token_stream:
        bytes = t.get_token_bits()
        download_bytes.extend(bytes)
        for b in bytes:
            download_str += chr(b)

    if (added_bytes > 0):
        extra_bytes = [0xff] * added_bytes
        download_bytes.extend(extra_bytes)
        for b in extra_bytes:
            download_str += chr(b)

    return download_bytes, download_str, download_type, version


# ********* Main and tests ********************************************


def test():
    global token_stream
    global err

    test_simp = ["incb %acc", "movb 12,@lil_count", "decw 1001/2", " # a comment line",
                  "   addb  $12 #do some adding",
                  "movb $'a', %acc", "branch :label1", "decw", "mulw 44", "subb $1000", "subw $-3000",
                  "movb $0, @big_count", "movw @last_reading, @store_reading"]

    test_jump = ["brne :f1", "bra $-1", "ret", ":f1", "dbnz :f2", "dsnz $10", "suba $10", ":f2"]
    test_stack = ["pushb $32", "pushw @count", "popb %acc", "popw 0x39", "pushw 43"]

    test_data = ["DATA home, 0", "DATB *, 0, 20, 'a'", ":start", "DATW intensity, *, *, -20",
                 'DATA buffer, *, 32, "This is an \"interesting\" string"']

    test_spec = ["LIMITS -1, -1, -1, 0, 0", "LIMITS 30, 10, 64, 0, 200",
                 "RESERVA 0, 16", "RESERVB 31,1", "RESERVW 0, 1"]

    test_device = ["DEVICE bad, 1", "DEVICE motor, 0", "DEVICE motor 13", "DEVICE motor 1 **",
                   "DEVICE digin, 1", "DEVICE motor 2, left_wheel", "DEVICE motor 4 right_wheel",
                   "DEVICE digout, 5", "DEVICE analogin 6"]
    test_begin_end = ["BEGIN WTF", "BEGIN PROGRAM 1", "BEGIN CONFIG", "BEGIN EVENT 34, 43,44",
                      "END PROGRAM CONFIG 0 0",
                      "BEGIN CONFIG 0,0", "BEGIN PROGRAM", "BEGIN EVENT %acc 0xfe 20/16"]

    test_other = ["stop 1", "COMMS", "COMMS 1 2", "FINISH 2",
                  ":start1", ":start2",
                  "stop", "FINISH", "COMMS 1024", "COMMS 400/16", "COMMS 0x400"]

    test_prog1 = ["COMMS 0x400", "LIMITS 20, 10, 64, 0, 200", 'RESERVB 0, 3', 'RESERVB 6, 2',
                  'RESERVW 0, 2', 'RESERVW 4,1', 'RESERVA 0, 5',
                  "BEGIN CONFIG 0,0", "BEGIN PROGRAM", "DATB count *", "DATB buffer *, 10",
                  "DATW temps *, 5", 'DATA message * * "This is fun but long - will it work?"',
                  "movb 5, @count", ":b1", "movb @count, %acc", "cmpb 0",
                  "bre :f1", "decb @count", "bra :b1",
                  ":f1", "movb 0, %acc", "END PROGRAM", "END CONFIG", "FINISH"]


    test_prog1 = ["COMMS 0x400", "LIMITS 200, 10, 64, 0, 200", 'RESERVB 0, 3', 'RESERVB 6, 2',
                  'RESERVW 0, 2', 'RESERVW 4,1', 'RESERVA 0, 5', 'DEVICE motor 1', 'DEVICE digin 2',
                  "BEGIN CONFIG 0,0", "BEGIN MAIN", "DATB count *", "DATB buffer *, 10",
                  "DATW temps *, 5", 'DATA message * * "This is fun but long - will it work?"',
                  '::top', 'DATB *, 0, 200',
                  "movb $5, @count", ":b1", "movb @count, %acc", "cmpb 0",
                  "bre :f1", "decb @count", "bra :b1",
                  ":f1", "movb 0, %acc", 'bra ::top',
                  "END MAIN",
                  "BEGIN EVENT %2C 1 1", ":f1", 'movw $-150, @temps', 'brne :f1',
                  'INSERT BINARY email.sig', 'bra :f1', 'END EVENT',
                  "END CONFIG", "FINISH"]

    test_prog2 = ['BEGIN FIRMWARE 0, 0, 0', 'BINB 0x10 20/16 255', 'BINB "Copyright"',
                  'INSERT BINARY email.sig', 'END FIRMWARE', 'FINISH']

    test_new_bad = ['bitset', 'bitset 3', 'bitset 3 2', 'bitset 29, f3'] # Bads
    test_new_good = ['BEGIN CONFIG 0,0', 'BEGIN MAIN',
                     'bitset 3 f3', 'bitset 0 ff',
                     'END MAIN', 'END CONFIG', 'FINISH']

    io.Out.DebugRaw("Starting test")

    token_stream = tokens.TokenStream()
    token_stream.clear()

    test_lines = []

    #test_lines.extend(test_simp)
    #test_lines.extend(test_jump)
    #test_lines.extend(test_stack)
    #test_lines.extend(test_data)
    #test_lines.extend(test_spec)
    #test_lines.extend(test_device)
    #test_lines.extend(test_begin_end)
    #test_lines.extend(test_other)
    #test_lines.extend(test_prog2)
    test_lines.extend(test_new_good)

    for t in test_lines:
        assem_line(t)

    token_stream.dump_tokens()
    io.Out.DebugDumpObjectRaw(token_stream, "TokenStream")

    token_analysis = tokens.TokenAnalyser(token_stream)
    token_analysis.map_all_variables()
    token_analysis.dump_variable_map()
    token_analysis.fixup_jumps()
    #token_analysis.fixup_sections()     # Add section headers including lengths
    token_analysis.fixup_jumps()        # Fixup globals which may change because of section headers
    token_analysis.fixup_crcs()         # Finally the crcs -- nothing will change now

    token_stream.dump_tokens()
    io.Out.DebugDumpObjectRaw(token_analysis, "Token_analyser")



from setuptools import setup
import os.path
import sys

setupdir = os.path.dirname(__file__)

requirements = []
for line in open(os.path.join(setupdir, "requirements.txt"), encoding="UTF-8"):
    if line.strip() and not line.startswith("#"):
        requirements.append(line)

setup(
    name="thonny-edison",
    version="1.0.0",
    description="A plug-in which adds Edison robot support for Thonny",
    long_description="""Plug-in for Thonny which adds Edison robot support.
 
More info: https://github.com/thonny/thonny-edison/""",
    url="https://github.com/thonny/thonny-edison/",
    author="Aivar Annamaa",
    license="GPL",
    classifiers=[
        "Environment :: MacOS X",
        "Environment :: Win32 (MS Windows)",
        "Environment :: X11 Applications",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: End Users/Desktop",
        "License :: Freeware",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Education",
        "Topic :: Software Development",
    ],
    keywords="IDE education programming robotics Edison",
    platforms=["Windows", "macOS", "Linux"],
    python_requires=">=3.5",
    package_data={
        "thonnycontrib.edison": ["res/*"],
        "thonnycontrib.backend.edison_backend": ["EdPy/*.py", "EdPy/*.json", "EdPy/LICENSE", "EdPy/lib/*.py"]
    },
    install_requires=requirements,
    py_modules=["Ed"],
    packages=["thonnycontrib.edison", "thonnycontrib.backend.edison_backend"],
)



#!/usr/bin/env python2
# * **************************************************************** **
# File: edpy_values.py
# Requires: Python 2.7+ (but not Python 3.0+)
# Note: For history, changes and dates for this file, consult git.
# Author: Brian Danilko, Likeable Software (brian@likeablesoftware.com)
# Copyright 2015-2017 Microbric Pty Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License (in the doc/licenses directory)
# for more details.
#
# * **************************************************************** */

"""Contains Ed.Py function signatures and constants used in
   parsing , optimising and compiling.
"""

# function signatures - note Ed is a module not a class!
# sig types: I - int, S - string const, V - integer list constant (vector)
#          : T - tunestr ref, L - list ref, O - object ref

# Spec version 1.4

signatures = {
    # Control
    "Ed.LeftLed": [('I', None)],
    "Ed.RightLed": [('I', None)],
    "Ed.ObstacleDetectionBeam": [('I', None)],
    "Ed.LineTrackerLed": [('I', None)],
    "Ed.SendIRData": [('I', None)],
    "Ed.StartCountDown": [('I', None), ('I', None)],   # time, units
    "Ed.TimeWait": [('I', None), ('I', None)],         # time, units
    "Ed.ResetDistance": [],

    # Music
    "Ed.PlayBeep": [],
    "Ed.PlayMyBeep": [('I', None)],                    # frequency
    "Ed.PlayTone": [('I', None), ('I', None)],         # frequency code, duration
    "Ed.PlayTune": [('T', None)],
    "Ed.ChangeTempo": [('I', None)],                  # new tempo

    # Movement
    "Ed.Drive": [('I', None), ('I', None), ('I', None)],       # direction, speed, distance
    "Ed.DriveLeftMotor": [('I', None), ('I', None), ('I', None)],
    "Ed.DriveRightMotor": [('I', None), ('I', None), ('I', None)],
    "Ed.SetDistance": [('I', None), ('I', None)],

    # Read data - return an int
    "Ed.ReadObstacleDetection": [],
    "Ed.ReadKeypad": [],
    "Ed.ReadClapSensor": [],
    "Ed.ReadLineState": [],
    "Ed.ReadLineChange": [],
    "Ed.ReadRemote": [],
    "Ed.ReadIRData": [],
    "Ed.ReadLeftLightLevel": [],
    "Ed.ReadRightLightLevel": [],
    "Ed.ReadLineTracker": [],
    "Ed.ReadCountDown": [('I', None)],
    "Ed.ReadMusicEnd": [],

    "Ed.ReadDriveLoad": [],               # OR of both sides
    "Ed.ReadDistance": [('I', None)],     # side - converts to inch or cm

    # NEW functions
    "Ed.ReadRandom": [],
    "Ed.ReadTuneError": [],

    #
    # -------------------------------------------------------------------------
    #
    # INLINE FUNCTIONS!
    #
    # NOTE -- some inline functions are created by the optimiser (and not the user)
    # They are still listed here so that the arguement types can be checked,
    # and return or not can be recorded.
    # That means that they could be used directly (by the user) -- that's not a bad thing, I think
    #
    # Any updates to these functions need to be changed in:
    # edpy_values.signatures, edpy_code.CODE,
    # compiler.SPECIALLY_HANDLED_FUNCTIONS, compiler.HandleSpecialCall
    #

    # Python normal functions
    "ord": [('S', 1)],
    "chr": [('I', None)],
    "len": [("VSTL", None)],    # can handle ListConst, StrConst, or TuneString/List refs
    "abs": [('I', None)],

    # Create special variables and objects. Also register event handlers
    "Ed.List1": [('I', None)],
    "Ed.List2": [('I', None), ('V', None)],
    "Ed.TuneString1": [('I', None)],
    "Ed.TuneString2": [('I', None), ('S', None)],
    "Ed.CreateObject": [('S', None)],
    "Ed.RegisterEventHandler": [('I', None), ('S', None)],

    # Low level access function
    "Ed.Init": [],
    "Ed.WriteModuleRegister8Bit": [('I', None), ('I', None), ('I', None)],
    "Ed.ReadModuleRegister8Bit": [('I', None), ('I', None)],
    "Ed.WriteModuleRegister16Bit": [('I', None), ('I', None), ('I', None)],
    "Ed.ReadModuleRegister16Bit": [('I', None), ('I', None)],

    "Ed.ClearModuleRegisterBit": [('I', None), ('I', None), ('I', None)],
    "Ed.SetModuleRegisterBit": [('I', None), ('I', None), ('I', None)],
    "Ed.AndModuleRegister8Bit": [('I', None), ('I', None), ('I', None)],
    "Ed.ObjectAddr": [('T', None)],

    # NEW simple motor functions (implemented in compiler)
    "Ed.SimpleDriveForwardRight": [],
    "Ed.SimpleDriveForwardLeft": [],
    "Ed.SimpleDriveStop": [],

    "Ed.SimpleDriveForward": [],
    "Ed.SimpleDriveBackward": [],
    "Ed.SimpleDriveBackwardRight": [],
    "Ed.SimpleDriveBackwardLeft": [],

    # optimisations of drive functions, all args constants and
    # unlimited distance set_comms
    "Ed.Drive_INLINE_UNLIMITED": [('I', None), ('I', None), ('I', None)],       # direction, speed, distance
    "Ed.DriveLeftMotor_INLINE_UNLIMITED": [('I', None), ('I', None), ('I', None)],
    "Ed.DriveRightMotor_INLINE_UNLIMITED": [('I', None), ('I', None), ('I', None)],
}


versionStatement = "VERSION 6, 0"

moduleStatements = [
    "DEVICE tracker, 0, LINE_TRACKER1",
    "DEVICE led, 1, Right_LED",
    "DEVICE motor-a, 3, Right_Motor",
    "DEVICE irrx, 5, IR_RECEIVER1",
    "DEVICE beeper, 6, SOUNDER1",
    "DEVICE irtx, 7, IR_TRANSMITTER1",
    "DEVICE motor-b, 8, Left_Motor",
    "DEVICE led, 11, Left_LED",
    ]

constants = {
    "Ed.ON": 1,
    "Ed.OFF": 0,

    "Ed.V1": 1,
    "Ed.V2": 2,

    "Ed.NOTE_A_6": 18181,        # 1760 Hz, 18181 count
    "Ed.NOTE_A_SHARP_6": 17167,  # 1864 Hz, 17167 count
    "Ed.NOTE_B_SHARP_6": 17167,  # 1864 Hz, 17167 count, NAME KEPT FOR COMPATIBILITY
    "Ed.NOTE_B_6": 16202,        # 1975 Hz, 16202 count
    "Ed.NOTE_C_7": 15289,        # 2093 Hz, 15289 count
    "Ed.NOTE_C_SHARP_7": 14433,  # 2217 Hz, 14433 count
    "Ed.NOTE_D_7": 13622,        # 2349 Hz, 13622 count
    "Ed.NOTE_D_SHARP_7": 12856,  # 2489 Hz, 12856 count
    "Ed.NOTE_E_7": 12135,        # 2637 Hz, 12135 count
    "Ed.NOTE_E_SHARP_7": 12135,  # 2637 Hz, 12135 count, NAME KEPT FOR COMPATIBILITY
    "Ed.NOTE_F_7": 11457,        # 2793 Hz, 11457 count
    "Ed.NOTE_F_SHARP_7": 10814,  # 2959 Hz, 10814 count
    "Ed.NOTE_G_7": 10207,        # 3135 Hz, 10207 count
    "Ed.NOTE_G_SHARP_7": 9632,   # 3322 Hz, 9632 count
    "Ed.NOTE_A_7": 9090,         # 3520 Hz, 9090 count
    "Ed.NOTE_A_SHARP_7": 8581,   # 3729 Hz, 8581 count
    "Ed.NOTE_B_SHARP_7": 8581,   # 3729 Hz, 8581 count, NAME KEPT FOR COMPATIBILITY
    "Ed.NOTE_B_7": 8099,         # 3951 Hz, 8099 count
    "Ed.NOTE_C_8": 7644,         # 4186 Hz, 7644 count
    "Ed.NOTE_REST": 0,

    # In milliseconds, using a whole note as 2 second
    "Ed.NOTE_SIXTEENTH": 125,
    "Ed.NOTE_EIGHT":     250,
    "Ed.NOTE_QUARTER":   500,
    "Ed.NOTE_HALF":      1000,
    "Ed.NOTE_WHOLE":     2000,

    "Ed.TEMPO_VERY_SLOW": 1000,
    "Ed.TEMPO_SLOW":      500,
    "Ed.TEMPO_MEDIUM":    250,
    "Ed.TEMPO_FAST":      70,
    "Ed.TEMPO_VERY_FAST": 1,


    # Motor directions
    "Ed.STOP": 0,

    # with distance
    "Ed.FORWARD": 1,
    "Ed.BACKWARD": 2,

    "Ed.DIR_COMPLEX_START": 3,

    # with degrees
    "Ed.FORWARD_RIGHT": 3,
    "Ed.BACKWARD_RIGHT": 4,
    "Ed.FORWARD_LEFT": 5,
    "Ed.BACKWARD_LEFT": 6,

    "Ed.DIR_SPIN_START": 7,

    "Ed.SPIN_RIGHT": 7,
    "Ed.SPIN_LEFT": 8,

    "Ed.SPEED_FULL": 0,
    "Ed.SPEED_1": 1,
    "Ed.SPEED_2": 2,
    "Ed.SPEED_3": 3,
    "Ed.SPEED_4": 4,
    "Ed.SPEED_5": 5,
    "Ed.SPEED_6": 6,
    "Ed.SPEED_7": 7,
    "Ed.SPEED_8": 8,
    "Ed.SPEED_9": 9,
    "Ed.SPEED_10": 10,

    "Ed.DISTANCE_UNLIMITED": 0,

    "Ed.MOTOR_LEFT":     0x00,
    "Ed.MOTOR_RIGHT":    0x01,

    # Internal values
    "Ed.MOTOR_FOR_CODE":     0x80,
    "Ed.MOTOR_BACK_CODE":    0x40,
    "Ed.MOTOR_DIST_CODE":    0x20,
    "Ed.MOTOR_FOR_DIST_CODE":  0xa0,
    "Ed.MOTOR_BACK_DIST_CODE": 0x60,
    "Ed.MOTOR_STOP_CODE":    0xc0,

    "Ed.OBSTACLE_NONE":     0x00,
    "Ed.OBSTACLE_DETECTED": 0x40,
    "Ed.OBSTACLE_LEFT":     0x20,
    "Ed.OBSTACLE_AHEAD":    0x10,
    "Ed.OBSTACLE_RIGHT":    0x08,
    "Ed.OBSTACLE_MASK":     0x78,
    "Ed.OBSTACLE_OTHER_MASK": 0x07,

    "Ed.LINE_ON_BLACK":      0x01,
    "Ed.LINE_ON_WHITE":      0x00,
    "Ed.LINE_MASK":          0x01,
    "Ed.LINE_CHANGE_MASK":   0x02,
    "Ed.LINE_CHANGE_BIT":    1,
    "Ed.LINE_CHANGE_MASK":   0x02,

    "Ed.KEYPAD_NONE":        0x00,
    "Ed.KEYPAD_TRIANGLE":    0x01,
    "Ed.KEYPAD_ROUND":       0x04,
    "Ed.KEYPAD_MASK":        0x0f,

    "Ed.CLAP_NOT_DETECTED":  0x00,
    "Ed.CLAP_DETECTED":      0x04,
    "Ed.CLAP_MASK":          0x04,
    "Ed.CLAP_DETECTED_BIT":  2,
    "Ed.DRIVE_STRAINED":     0x01,
    "Ed.DRIVE_NO_STRAIN":    0x00,
    "Ed.MUSIC_FINISHED":     0x01,
    "Ed.MUSIC_NOT_FINISHED": 0x00,
    "Ed.TUNE_NO_ERROR":      0x00,
    "Ed.TUNE_ERROR":         0x01,

    "Ed.REMOTE_CODE_0": 0,
    "Ed.REMOTE_CODE_1": 1,
    "Ed.REMOTE_CODE_2": 2,
    "Ed.REMOTE_CODE_3": 3,
    "Ed.REMOTE_CODE_4": 4,
    "Ed.REMOTE_CODE_5": 5,
    "Ed.REMOTE_CODE_6": 6,
    "Ed.REMOTE_CODE_7": 7,

    "Ed.REMOTE_CODE_NONE" : 255,

    "Ed.EVENT_TIMER_FINISHED": 0,
    "Ed.EVENT_REMOTE_CODE": 1,
    "Ed.EVENT_IR_DATA": 2,
    "Ed.EVENT_CLAP_DETECTED": 3,
    "Ed.EVENT_OBSTACLE_ANY": 4,
    "Ed.EVENT_OBSTACLE_LEFT": 5,
    "Ed.EVENT_OBSTACLE_RIGHT": 6,
    "Ed.EVENT_OBSTACLE_AHEAD": 7,
    "Ed.EVENT_DRIVE_STRAIN": 8,
    "Ed.EVENT_KEYPAD_TRIANGLE": 9,
    "Ed.EVENT_KEYPAD_ROUND": 10,
    "Ed.EVENT_LINE_TRACKER_ON_WHITE": 11,
    "Ed.EVENT_LINE_TRACKER_ON_BLACK": 12,
    "Ed.EVENT_LINE_TRACKER_SURFACE_CHANGE": 13,
    "Ed.EVENT_TUNE_FINISHED": 14,

    "Ed.EVENT_LAST_EVENT" : 14,

    "Ed.CM":     0x00,
    "Ed.INCH":   0x01,
    "Ed.TIME":   0x02,

    "Ed.TIME_SECONDS":       0x00,
    "Ed.TIME_MILLISECONDS":  0x01,


    # Used for the Ed.Py functions and low level access

    "Ed.MODULE_LINE_TRACKER": 0,
    "Ed.MODULE_RIGHT_LED": 1,
    "Ed.MODULE_RIGHT_MOTOR": 3,
    "Ed.MODULE_IR_RX": 5,
    "Ed.MODULE_BEEPER": 6,
    "Ed.MODULE_IR_TX": 7,
    "Ed.MODULE_LEFT_MOTOR": 8,
    "Ed.MODULE_LEFT_LED": 11,

    "Ed.MODULE_INDEX": 12,
    "Ed.MODULE_DEVICES": 13,
    "Ed.MODULE_TIMERS": 14,
    "Ed.MODULE_CPU": 15,

    # Line Tracker
    "Ed.REG_LT_STATUS_8": 0,
    "Ed.REG_LT_POWER_8": 1,
    "Ed.REG_LT_LEVEL_16": 2,

    # LEDs
    "Ed.REG_LED_STATUS_8": 0,
    "Ed.REG_LED_OUTPUT_8": 1,
    "Ed.REG_LED_LEVEL_16": 2,

    # Motors
    "Ed.REG_MOTOR_STATUS_8": 0,
    "Ed.REG_MOTOR_CONTROL_8": 1,
    "Ed.REG_MOTOR_DISTANCE_16": 2,

    # IR RX
    "Ed.REG_IRRX_STATUS_8": 0,
    "Ed.REG_IRRX_ACTION_8": 1,
    "Ed.REG_IRRX_CHECK_INDEX_8": 2,
    "Ed.REG_IRRX_MATCH_INDEX_8": 3,
    "Ed.REG_IRRX_RCV_CHAR_8": 4,

    # BEEPER
    "Ed.REG_BEEP_STATUS_8": 0,
    "Ed.REG_BEEP_ACTION_8": 1,
    "Ed.REG_BEEP_FREQ_16": 2,
    "Ed.REG_BEEP_DURATION_16": 4,
    "Ed.REG_BEEP_TUNE_CODE_8": 6,
    "Ed.REG_BEEP_TUNE_STRING_8": 7,
    "Ed.REG_BEEP_TUNE_TEMPO_16": 8,

    # IR TX
    "Ed.REG_IRTX_ACTION_8": 0,
    "Ed.REG_IRTX_CHAR_8": 1,

    # INDEX

    # DEVICES
    "Ed.REG_DEV_STATUS_8": 0,
    "Ed.REG_DEV_ACTION_8": 1,
    "Ed.REG_DEV_RANDOM_8": 0x0c,
    "Ed.REG_DEV_BUTTON_8": 0x0d,

    # TIMERS
    "Ed.REG_TIMER_STATUS_8": 0,
    "Ed.REG_TIMER_ACTION_8": 1,
    "Ed.REG_TIMER_PAUSE_16": 2,
    "Ed.REG_TIMER_ONE_SHOT_16": 4,
    "Ed.REG_TIMER_SYS_TIME_16": 6,

    # CPU
}

# Values MUST be set and can only be set once! Must be in the main code (not a function).
variables = {
    # name : tuple of allowed values
    "Ed.EdisonVersion": (constants["Ed.V1"], constants["Ed.V2"]),
    "Ed.DistanceUnits": (constants["Ed.CM"], constants["Ed.INCH"],
                         constants["Ed.TIME"]),
    "Ed.Tempo": (constants["Ed.TEMPO_VERY_SLOW"], constants["Ed.TEMPO_SLOW"],
                 constants["Ed.TEMPO_MEDIUM"], constants["Ed.TEMPO_FAST"],
                 constants["Ed.TEMPO_VERY_FAST"]),
}

notAvailableFunctions = {
    1: ("Ed.ResetDistance", "Ed.SetDistance", "Ed.ReadDistance"),   # Ed.V1
    2: (),                                                          # Ed.V2
}

# QUESTIONS FOR BILL
#
# 1. Does the Devices.button register have to be cleared? Does it have a bitmask? Can more then
#    one button be pressed? What are the values and how do they compare to the shapes?
#
